"""Reader-Baseclass."""

from __future__ import annotations

import contextlib
import errno
import logging
import math
import os
from datetime import datetime
from itertools import product
from pathlib import Path
from types import MappingProxyType
from typing import TYPE_CHECKING
from typing import Any

import h5py
import numpy as np
import yaml
from pydantic import validate_call
from tqdm import trange
from typing_extensions import Self
from typing_extensions import deprecated

from .config import config
from .data_models.base.calibration import CalibrationPair
from .data_models.base.calibration import CalibrationSeries
from .data_models.base.timezone import local_tz
from .data_models.content.energy_environment import EnergyDType
from .decoder_waveform import Uart

if TYPE_CHECKING:
    from collections.abc import Generator
    from collections.abc import Mapping
    from collections.abc import Sequence
    from types import TracebackType


class Reader:
    """Sequentially Reads shepherd-data from HDF5 file.

    Args:
    ----
        file_path: Path of hdf5 file containing shepherd data with iv-samples, iv-curves or isc&voc
        verbose: more debug-info during usage, 'None' skips the setter

    """

    CHUNK_SAMPLES_N: int = 10_000

    MODE_TO_DTYPE: Mapping[str, Sequence[EnergyDType]] = MappingProxyType(
        {
            "harvester": (
                EnergyDType.ivsample,
                EnergyDType.ivcurve,
                EnergyDType.isc_voc,
            ),
            "emulator": (EnergyDType.ivsample,),
        }
    )

    @validate_call
    def __init__(
        self,
        file_path: Path,
        *,
        verbose: bool = True,
    ) -> None:
        self.file_path: Path = file_path.resolve()

        if not hasattr(self, "_logger"):
            self._logger: logging.Logger = logging.getLogger("SHPCore.Reader")
        if verbose is not None:
            self._logger.setLevel(logging.DEBUG if verbose else logging.INFO)

        if not hasattr(self, "samplerate_sps"):
            self.samplerate_sps: int = config.SAMPLERATE_SPS
        self.sample_interval_ns: int = round(10**9 // self.samplerate_sps)
        self.sample_interval_s: float = 1 / self.samplerate_sps

        self.max_elements: int = 40 * self.samplerate_sps
        # ⤷ per iteration (40s full res, < 200 MB RAM use)

        # init stats
        self.runtime_s: float = 0
        self.samples_n: int = 0
        self.chunks_n: int = 0
        self.file_size: int = 0
        self.data_rate: float = 0

        # open file (if not already done by writer)
        self._reader_opened: bool = False
        if not hasattr(self, "h5file"):
            if not isinstance(self.file_path, Path):
                raise TypeError("Provide a valid Path-Object to Reader!")
            if not (self.file_path.exists() and self.file_path.is_file()):
                raise FileNotFoundError(
                    errno.ENOENT, os.strerror(errno.ENOENT), self.file_path.name
                )

            try:
                self.h5file = h5py.File(self.file_path, "r")  # = readonly
                self._reader_opened = True
            except OSError as _xcp:
                msg = f"Unable to open HDF5-File '{self.file_path.name}'"
                raise TypeError(msg) from _xcp

            if self.is_valid():
                self._logger.debug("File is available now")
            else:
                self._logger.error(
                    "[FileValidation] File is faulty! "
                    "Will try to open but there might be dragons, for '%s'",
                    self.file_path.name,
                )

        if not isinstance(self.h5file, h5py.File):
            msg = (f"Type of opened file is not h5py.File, for {self.file_path.name}",)
            raise TypeError(msg)

        self.ds_time: h5py.Dataset = self.h5file["data"]["time"]
        self.ds_voltage: h5py.Dataset = self.h5file["data"]["voltage"]
        self.ds_current: h5py.Dataset = self.h5file["data"]["current"]

        # retrieve cal-data
        if not hasattr(self, "_cal"):
            cal_dict = CalibrationSeries().model_dump()
            for ds, param in product(["current", "voltage", "time"], ["gain", "offset"]):
                try:
                    cal_dict[ds][param] = self.h5file["data"][ds].attrs[param]
                except KeyError:  # noqa: PERF203
                    self._logger.debug("Cal-Param '%s' for dataset '%s' not found!", param, ds)
            self._cal = CalibrationSeries(**cal_dict)

        self._refresh_file_stats()

        if file_path is not None:
            # file opened by this reader
            self._logger.debug(
                "Reading data from '%s'\n"
                "\t- runtime %.1f s\n"
                "\t- mode = %s\n"
                "\t- window_size = %d\n"
                "\t- size = %.3f MiB\n"
                "\t- rate = %.0f KiB/s",
                self.file_path.name,
                self.runtime_s,
                self.get_mode(),
                self.get_window_samples(),
                self.file_size / 2**20,
                self.data_rate / 2**10,
            )

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        typ: type[BaseException] | None = None,
        exc: BaseException | None = None,
        tb: TracebackType | None = None,
        extra_arg: int = 0,
    ) -> None:
        if self._reader_opened:
            self.h5file.close()

    def __repr__(self) -> str:
        return yaml.safe_dump(
            self.get_metadata(minimal=True), default_flow_style=False, sort_keys=False
        )

    def _refresh_file_stats(self) -> None:
        """Update internal states, helpful after resampling or other changes in data-group."""
        self.h5file.flush()
        self.samples_n = min(
            self.ds_time.shape[0], self.ds_current.shape[0], self.ds_voltage.shape[0]
        )
        duration_raw = (
            (int(self.ds_time[self.samples_n - 1]) - int(self.ds_time[0]))
            if self.samples_n > 0
            else 0
        )
        # above's typecasting prevents overflow in u64-format
        if (self.samples_n > 0) and (duration_raw > 0):
            # this assumes iso-chronous sampling, TODO: not the best choice?
            duration_s = self._cal.time.raw_to_si(duration_raw)
            self.sample_interval_s = duration_s / self.samples_n
            self.sample_interval_ns = round(10**9 * self.sample_interval_s)
            self.samplerate_sps = max(round((self.samples_n - 1) / duration_s), 1)
        self.runtime_s = round(self.samples_n / self.samplerate_sps, 1)
        self.chunks_n = int(self.samples_n // self.CHUNK_SAMPLES_N)
        if isinstance(self.file_path, Path):
            self.file_size = self.file_path.stat().st_size
        else:
            self.file_size = 0
        self.data_rate = self.file_size / self.runtime_s if self.runtime_s > 0 else 0

    def read(
        self,
        start_n: int = 0,
        end_n: int | None = None,
        n_samples_per_chunk: int | None = None,
        *,
        is_raw: bool = False,
        omit_timestamps: bool = False,
    ) -> Generator[tuple, None, None]:
        """Read the specified range of chunks from the hdf5 file.

        Generator - can be configured on first call

        Args:
        ----
            :param start_n: (int) Index of first chunk to be read
            :param end_n: (int) Index of last chunk to be read
            :param n_samples_per_chunk: (int) allows changing
            :param is_raw: (bool) output original data, not transformed to SI-Units
            :param omit_timestamps: (bool) optimize reading if timestamp is never used
        Yields: chunks between start and end (tuple with time, voltage, current)

        """
        if n_samples_per_chunk is None:
            n_samples_per_chunk = self.CHUNK_SAMPLES_N
        end_max = int(self.samples_n // n_samples_per_chunk)
        end_n = end_max if end_n is None else min(end_n, end_max)
        self._logger.debug("Reading chunk %d to %d from source-file", start_n, end_n)
        _raw = is_raw
        _wts = not omit_timestamps

        for i in range(start_n, end_n):
            idx_start = i * n_samples_per_chunk
            idx_end = idx_start + n_samples_per_chunk
            if _raw:
                yield (
                    self.ds_time[idx_start:idx_end] if _wts else None,
                    self.ds_voltage[idx_start:idx_end],
                    self.ds_current[idx_start:idx_end],
                )
            else:
                yield (
                    self._cal.time.raw_to_si(self.ds_time[idx_start:idx_end]) if _wts else None,
                    self._cal.voltage.raw_to_si(self.ds_voltage[idx_start:idx_end]),
                    self._cal.current.raw_to_si(self.ds_current[idx_start:idx_end]),
                )

    @deprecated("use .read() instead")
    def read_buffers(
        self,
        start_n: int = 0,
        end_n: int | None = None,
        n_samples_per_buffer: int | None = None,
        *,
        is_raw: bool = False,
        omit_ts: bool = False,
    ) -> Generator[tuple, None, None]:
        return self.read(
            start_n=start_n,
            end_n=end_n,
            n_samples_per_chunk=n_samples_per_buffer,
            is_raw=is_raw,
            omit_timestamps=omit_ts,
        )

    def get_time_start(self) -> datetime | None:
        if self.samples_n < 1:
            return None
        return datetime.fromtimestamp(self._cal.time.raw_to_si(self.ds_time[0]), tz=local_tz())

    def get_calibration_data(self) -> CalibrationSeries:
        """Read calibration-data from hdf5 file.

        :return: Calibration data as CalibrationSeries object
        """
        return self._cal

    def get_window_samples(self) -> int:
        if "window_samples" in self.h5file["data"].attrs:
            return int(self.h5file["data"].attrs["window_samples"])
        return 0

    def get_mode(self) -> str:
        if "mode" in self.h5file.attrs:
            return self.h5file.attrs["mode"]
        return ""

    def get_config(self) -> dict:
        if "config" in self.h5file["data"].attrs:
            return yaml.safe_load(self.h5file["data"].attrs["config"])
        return {}

    def get_hostname(self) -> str:
        if "hostname" in self.h5file.attrs:
            return self.h5file.attrs["hostname"]
        return "unknown"

    def get_datatype(self) -> EnergyDType | None:
        try:
            if "datatype" in self.h5file["data"].attrs:
                return EnergyDType[self.h5file["data"].attrs["datatype"]]
        except KeyError:
            return None
        except ValueError:
            return None
        else:
            return None

    def get_voltage_step(self) -> float | None:
        """Informs about the voltage step (in volts) used during harvesting the ivcurve.

        Options for figuring out the real step:
        - look into config (if available)
        - analyze recorded data for most often used delta
        - calculate with 'steps_n * (1 + wait_cycles)' (done for calculating window_size)
        """
        voltage_step: float | None = (
            self.get_config().get("virtual_harvester", {}).get("voltage_step_mV", None)
        )
        if voltage_step is not None:  # convert mV to V
            voltage_step = 1e-3 * voltage_step
        if voltage_step is None:
            dsv = self._cal.voltage.raw_to_si(self.ds_voltage[0:2000])
            diffs_np = np.unique(dsv[1:] - dsv[0:-1], return_counts=False)
            diffs_ls = [_e for _e in list(np.array(diffs_np)) if _e > 0]
            # static voltages have 0 steps, so
            if len(diffs_ls) == 0:
                self._logger.warning("Voltage-Step could not be determined from source-material")
                return None  # or is 0 better? that may provoke div0
            voltage_step = min(diffs_ls)
        return voltage_step

    def get_hrv_config(self) -> dict:
        """Essential info for harvester.

        :return: config-dict directly for vHarvester to be used during emulation
        """
        return {
            "datatype": self.get_datatype(),
            "window_samples": self.get_window_samples(),
            "voltage_step_V": self.get_voltage_step(),
        }

    def is_valid(self) -> bool:
        """Check file for plausibility, validity / correctness.

        :return: state of validity
        """
        # hard criteria
        if "data" not in self.h5file:
            self._logger.error(
                "[FileValidation] root data-group not found in '%s'",
                self.file_path.name,
            )
            return False
        for attr in ["mode"]:
            if attr not in self.h5file.attrs:
                self._logger.error(
                    "[FileValidation] attribute '%s' not found in '%s'",
                    attr,
                    self.file_path.name,
                )
                return False
            if self.h5file.attrs["mode"] not in self.MODE_TO_DTYPE:
                self._logger.error(
                    "[FileValidation] unsupported mode '%s' in '%s'",
                    attr,
                    self.file_path.name,
                )
                return False
        for attr in ["window_samples", "datatype"]:
            if attr not in self.h5file["data"].attrs:
                self._logger.error(
                    "[FileValidationError] attribute '%s' not found in data-group in '%s'",
                    attr,
                    self.file_path.name,
                )
                return False
        for dset in ["time", "current", "voltage"]:
            if dset not in self.h5file["data"]:
                self._logger.error(
                    "[FileValidation] dataset '%s' not found in '%s'",
                    dset,
                    self.file_path.name,
                )
                return False
            for attr in ["gain", "offset"]:
                if attr not in self.h5file["data"][dset].attrs:
                    self._logger.error(
                        "[FileValidation] attribute '%s' not found in dataset '%s' in '%s'",
                        attr,
                        dset,
                        self.file_path.name,
                    )
                    return False
        if self.get_datatype() not in self.MODE_TO_DTYPE[self.get_mode()]:
            self._logger.error(
                "[FileValidation] unsupported type '%s' for mode '%s'  in '%s'",
                self.get_datatype(),
                self.get_mode(),
                self.file_path.name,
            )
            return False

        if self.get_datatype() == EnergyDType.ivcurve and self.get_window_samples() < 1:
            self._logger.error(
                "[FileValidation] window size / samples is < 1 "
                "-> invalid for ivcurve-datatype, in '%s'",
                self.file_path.name,
            )
            return False

        # soft-criteria:
        if self.get_datatype() != EnergyDType.ivcurve and self.get_window_samples() > 0:
            self._logger.warning(
                "[FileValidation] window size / samples is > 0 despite "
                "not using the ivcurve-datatype, in '%s'",
                self.file_path.name,
            )
        # same length of datasets:
        samples_n = self.h5file["data"]["time"].shape[0]
        for dset in ["voltage", "current"]:
            ds_size = self.h5file["data"][dset].shape[0]
            if ds_size != samples_n:
                self._logger.warning(
                    "[FileValidation] dataset '%s' has different size (=%d), "
                    "compared to time (=%d), in '%s'",
                    dset,
                    ds_size,
                    samples_n,
                    self.file_path.name,
                )
        # dataset-length should be multiple of chunk-size
        remaining_size = samples_n % self.CHUNK_SAMPLES_N
        if remaining_size != 0:
            self._logger.warning(
                "[FileValidation] datasets are not aligned with chunk-size in '%s'",
                self.file_path.name,
            )
        # check compression
        for dset in ["time", "current", "voltage"]:
            comp = self.h5file["data"][dset].compression
            opts = self.h5file["data"][dset].compression_opts
            if comp not in {None, "gzip", "lzf"}:
                self._logger.warning(
                    "[FileValidation] unsupported compression found "
                    "(%s != None, lzf, gzip) in '%s'",
                    comp,
                    self.file_path.name,
                )
            if (comp == "gzip") and (opts is not None) and (int(opts) > 1):
                self._logger.warning(
                    "[FileValidation] gzip compression is too high (%d > 1) for BBone in '%s'",
                    opts,
                    self.file_path.name,
                )
        # host-name
        if self.get_hostname() == "unknown":
            self._logger.warning(
                "[FileValidation] Hostname was not set in '%s'", self.file_path.name
            )
        # errors during execution
        _err = self.count_errors_in_log()
        if _err > 0:
            self._logger.warning(
                "[FileValidation] Sheep reported %d errors during execution -> check logs in '%s'",
                _err,
                self.file_path.name,
            )
        return True

    def __getitem__(self, key: str) -> Any:
        """Query attribute or (if none found) a handle for a group or dataset (if found).

        :param key: attribute, group, dataset
        :return: value of that key, or handle of object
        """
        if key in self.h5file.attrs:
            return self.h5file.attrs.__getitem__(key)
        if key in self.h5file:
            return self.h5file.__getitem__(key)
        raise KeyError

    def energy(self) -> float:
        """Determine the recorded energy of the trace.

        # multiprocessing: https://stackoverflow.com/a/71898911
        # -> failed with multiprocessing.pool and pathos.multiprocessing.ProcessPool.

        :return: sampled energy in Ws (watt-seconds)
        """
        iterations = math.ceil(self.samples_n / self.max_elements)
        job_iter = trange(
            0,
            self.samples_n,
            self.max_elements,
            desc="energy",
            leave=False,
            disable=iterations < 8,
        )

        def _calc_energy(idx_start: int) -> float:
            idx_stop = min(idx_start + self.max_elements, self.samples_n)
            vol_v = self._cal.voltage.raw_to_si(self.ds_voltage[idx_start:idx_stop])
            cur_a = self._cal.current.raw_to_si(self.ds_current[idx_start:idx_stop])
            return (vol_v[:] * cur_a[:]).sum() * self.sample_interval_s

        energy_ws = [_calc_energy(i) for i in job_iter]
        return float(sum(energy_ws))

    def _dset_statistics(
        self, dset: h5py.Dataset, cal: CalibrationPair | None = None
    ) -> dict[str, float]:
        """Create basic stats for a provided dataset.

        :param dset: dataset to evaluate
        :param cal: calibration (if wanted)
        :return: dict with entries for mean, min, max, std
        """
        si_converted = True
        if not isinstance(cal, CalibrationPair):
            if "gain" in dset.attrs and "offset" in dset.attrs:
                cal = CalibrationPair(gain=dset.attrs["gain"], offset=dset.attrs["offset"])
            else:
                cal = CalibrationPair(gain=1)
                si_converted = False
        iterations = math.ceil(dset.shape[0] / self.max_elements)
        job_iter = trange(
            0,
            dset.shape[0],
            self.max_elements,
            desc=f"{dset.name}-stats",
            leave=False,
            disable=iterations < 8,
        )

        def _calc_statistics(data: np.ndarray) -> list:
            return [np.mean(data), np.min(data), np.max(data), np.std(data)]

        stats_list = [
            _calc_statistics(cal.raw_to_si(dset[i : i + self.max_elements])) for i in job_iter
        ]
        if len(stats_list) < 1:
            return {}
        stats_nd = np.stack(stats_list)
        stats: dict[str, float] = {
            # TODO: wrong calculation for ndim-datasets with n>1
            "mean": float(stats_nd[:, 0].mean()),
            "min": float(stats_nd[:, 1].min()),
            "max": float(stats_nd[:, 2].max()),
            "std": float(stats_nd[:, 3].mean()),  # TODO: sooo wrong :)
            "si_converted": si_converted,
        }
        return stats

    def _data_timediffs(self) -> list[float]:
        """Calculate list of unique time-deltas [s] between chunks.

        Optimized version that only looks at the start of each chunk.
        Timestamps get converted to signed (it still fits > 100 years)
        to allow calculating negative diffs.

        :return: list of (unique) time-deltas between chunks [s]
        """
        iterations = math.ceil(self.samples_n / self.max_elements)
        job_iter = trange(
            0,
            self.samples_n,
            self.max_elements,
            desc="timediff",
            leave=False,
            disable=iterations < 8,
        )

        def calc_timediffs(idx_start: int) -> list:
            ds_time = self.ds_time[
                idx_start : (idx_start + self.max_elements) : self.CHUNK_SAMPLES_N
            ].astype(np.int64)
            diffs_np = np.unique(ds_time[1:] - ds_time[0:-1], return_counts=False)
            return list(np.array(diffs_np))

        diffs_ll = [calc_timediffs(i) for i in job_iter]
        diffs = {
            round(self._cal.time.raw_to_si(j) / self.CHUNK_SAMPLES_N, 6)
            for i in diffs_ll
            for j in i
        }
        return list(diffs)

    def check_timediffs(self) -> bool:
        """Validate equal time-deltas.

        Unexpected time-jumps hint at a corrupted file or faulty measurement.

        :return: True if OK
        """
        diffs = self._data_timediffs()
        if len(diffs) > 1:
            self._logger.warning(
                "Time-jumps detected -> expected equal steps, but got: %s s", diffs
            )
        return (len(diffs) <= 1) and diffs[0] == round(0.1 / self.CHUNK_SAMPLES_N, 6)

    def count_errors_in_log(self, group_name: str = "sheep", min_level: int = 40) -> int:
        if group_name not in self.h5file:
            return 0
        if "level" not in self.h5file[group_name]:
            return 0
        _lvl = self.h5file[group_name]["level"]
        if _lvl.shape[0] < 1:
            return 0
        _items = [1 for _x in _lvl[:] if _x >= min_level]
        return len(_items)

    def get_metadata(
        self,
        node: h5py.Dataset | h5py.Group | None = None,
        *,
        minimal: bool = False,
    ) -> dict[str, dict]:
        """Recursive FN to capture the structure of the file.

        :param node: starting node, leave free to go through whole file
        :param minimal: just provide a bare tree (much faster)
        :return: structure of that node with everything inside it
        """
        if node is None:
            self._refresh_file_stats()
            return self.get_metadata(self.h5file, minimal=minimal)

        metadata: dict[str, dict] = {}
        if isinstance(node, h5py.Dataset) and not minimal:
            metadata["_dataset_info"] = {
                "datatype": str(node.dtype),
                "shape": str(node.shape),
                "chunks": str(node.chunks),
                "compression": str(node.compression),
                "compression_opts": str(node.compression_opts),
            }
            if node.name == "/data/time":
                metadata["_dataset_info"]["time_diffs_s"] = self._data_timediffs()
                # TODO: already convert to str to calm the typechecker?
                #  or construct a pydantic-class
            elif "int" in str(node.dtype):
                metadata["_dataset_info"]["statistics"] = self._dset_statistics(node)
                # TODO: put this into metadata["_dataset_statistics"] ??
        for attr in node.attrs:
            attr_value = node.attrs[attr]
            if isinstance(attr_value, str):
                with contextlib.suppress(yaml.YAMLError):
                    attr_value = yaml.safe_load(attr_value)
            elif "int" in str(type(attr_value)):
                # TODO: why not isinstance? can it be list[int] other complex type?
                attr_value = int(attr_value)
            else:
                attr_value = float(attr_value)
            metadata[attr] = attr_value
        if isinstance(node, h5py.Group):
            if node.name == "/data" and not minimal:
                metadata["_group_info"] = {
                    "energy_Ws": self.energy(),
                    "runtime_s": round(self.runtime_s, 1),
                    "data_rate_KiB_s": round(self.data_rate / 2**10),
                    "file_size_MiB": round(self.file_size / 2**20, 3),
                    "valid": self.is_valid(),
                }
            for item in node:
                metadata[item] = self.get_metadata(node[item], minimal=minimal)

        return metadata

    def save_metadata(self, node: h5py.Dataset | h5py.Group | None = None) -> dict:
        """Get structure of file and dump content to yaml-file with same name as original.

        :param node: starting node, leave free to go through whole file
        :return: structure of that node with everything inside it
        """
        if isinstance(self.file_path, Path):
            yaml_path = Path(self.file_path).resolve().with_suffix(".yaml")
            if yaml_path.exists():
                self._logger.info("File already exists, will skip '%s'", yaml_path.name)
                return {}
            metadata = self.get_metadata(node)  # {"h5root": self.get_metadata(self.h5file)}
            with yaml_path.open("w", encoding="utf-8-sig") as yfd:
                yaml.safe_dump(metadata, yfd, default_flow_style=False, sort_keys=False)
        else:
            metadata = {}
        return metadata

    def get_gpio_pin_num(self, name: str) -> int | None:
        # reverse lookup in a 2D-dict: key1 are pin_num, key2 are descriptor-names
        if "gpio" not in self.h5file:
            return None
        descriptions = yaml.safe_load(self.h5file["gpio"]["value"].attrs["description"])
        for desc_name, desc in descriptions.items():
            if name in desc["name"]:
                return int(desc_name)
        return None

    @staticmethod
    def get_filter_for_redundant_states(data: np.ndarray) -> np.ndarray:
        """Input is 1D state-vector, kep only first from identical & sequential states.

        Algo: create an offset-by-one vector and compare against original.
        """
        if len(data.shape) > 1:
            raise ValueError("Array must be 1D")
        data_1 = np.concatenate(([not data[0]], data[:-1]))
        return data != data_1

    def gpio_to_waveforms(self, name: str | None = None) -> dict:
        waveforms: dict[str, np.ndarray] = {}
        if "gpio" not in self.h5file:
            return waveforms

        gpio_ts = self.h5file["gpio"]["time"]
        gpio_vs = self.h5file["gpio"]["value"]

        if name is None:
            descriptions = yaml.safe_load(self.h5file["gpio"]["value"].attrs["description"])
            pin_dict = {value["name"]: key for key, value in descriptions.items()}
        else:
            pin_dict = {name: self.get_gpio_pin_num(name)}

        for pin_name, pin_num in pin_dict.items():
            gpio_ps = (gpio_vs[:] & (0b1 << pin_num)) > 0
            gpio_f = self.get_filter_for_redundant_states(gpio_ps)
            waveforms[pin_name] = np.column_stack((gpio_ts[gpio_f], gpio_ps[gpio_f]))
            self._logger.debug(
                "GPIO '%s' has %d state-changes (includes initial state)",
                pin_name,
                sum(gpio_f),
            )
        return waveforms

    def waveform_to_csv(self, pin_name: str, pin_wf: np.ndarray, separator: str = ",") -> None:
        path_csv = self.file_path.with_suffix(f".waveform.{pin_name}.csv")
        if path_csv.exists():
            self._logger.info("File already exists, will skip '%s'", path_csv.name)
            return
        with path_csv.open("w") as csv:
            csv.write(f"timestamp [s],{pin_name}\n")
            for row in pin_wf:
                csv.write(f"{row[0] / 1e9}{separator}{int(row[1])}\n")

    def gpio_to_uart(self) -> np.ndarray | None:
        wfs = self.gpio_to_waveforms("uart")
        if len(wfs) < 1:
            return None
        pin_name, pin_wf = wfs.popitem()

        write_to_file = False
        if write_to_file:
            self.waveform_to_csv(pin_name, pin_wf)

        gpio_wf = pin_wf.astype(float)
        gpio_wf[:, 0] = gpio_wf[:, 0] / 1e9

        try:
            return Uart(gpio_wf).get_lines()
        except TypeError:
            self._logger.error("TypeError: Extracting UART from GPIO failed - will skip file.")
            return None
        except ValueError:
            self._logger.error("ValueError: Extracting UART from GPIO failed - will skip file.")
            return None



from shepherd_core.data_models.content.firmware import Firmware
from shepherd_core.data_models.testbed.cape import Cape
from shepherd_core.data_models.testbed.gpio import GPIO
from shepherd_core.data_models.testbed.mcu import MCU
from shepherd_core.data_models.testbed.observer import Observer
from shepherd_core.data_models.testbed.target import Target
from shepherd_core.data_models.testbed.testbed import Testbed as TasteBad
from shepherd_core.testbed_client.fixtures import Fixtures

# ⤷ TasteBad avoids pytest-warning


def test_testbed_fixture_cape() -> None:
    for fix in Fixtures()["Cape"]:
        Cape(name=fix["name"])
        Cape(id=fix["id"])


def test_testbed_fixture_gpio() -> None:
    for fix in Fixtures()["GPIO"]:
        GPIO(name=fix["name"])
        GPIO(id=fix["id"])


def test_testbed_fixture_mcu() -> None:
    for fix in Fixtures()["MCU"]:
        MCU(name=fix["name"])
        mcu = MCU(id=fix["id"])
        Firmware(name=mcu.fw_name_default)


def test_testbed_fixture_observer() -> None:
    for fix in Fixtures()["Observer"]:
        print(f"selecting {fix['name']} - {fix['id']}")
        Observer(name=fix["name"])
        Observer(id=fix["id"])


def test_testbed_fixture_target() -> None:
    for fix in Fixtures()["Target"]:
        Target(name=fix["name"])
        Target(id=fix["id"])


def test_testbed_fixture_tb() -> None:
    for fix in Fixtures()["Testbed"]:
        TasteBad(name=fix["name"])
        TasteBad(id=fix["id"])



"""This example manipulates elf-files and embeds firmware into the data-model.

Note: make sure to have installed
- shepherd-core[elf] via pip
- build-essential or binutils-$ARCH installed.
"""

import shutil
from pathlib import Path

from shepherd_core.data_models import Firmware
from shepherd_core.data_models import FirmwareDType

from shepherd_core import fw_tools

path_src = Path(__file__).parent.parent / "tests/fw_tools/build_msp.elf"
path_elf = Path(__file__).with_name("firmware_msp.elf")

# make local copy to play with
shutil.copy(path_src, path_elf)

print(f"UID old = 0x{fw_tools.read_uid(path_elf):X}")
fw_tools.modify_uid(path_elf, 0xCAFE)
print(f"UID new = 0x{fw_tools.read_uid(path_elf):X}")

path_hex = fw_tools.elf_to_hex(path_elf)

# just PoC - there is an easier way to generate data-model, see other fw-example
b64 = fw_tools.file_to_base64(path_elf)

fw = Firmware(
    name="msp_deep_sleep",
    data=b64,
    data_type=FirmwareDType.base64_elf,
    mcu={"name": "MSP430FR"},
    owner="example",
    group="test",
)

# for completion also generate hex for nrf
path_src = Path(__file__).parent.parent / "tests/fw_tools/build_nrf.elf"
path_elf = Path(__file__).with_name("firmware_nrf.elf")
shutil.copy(path_src, path_elf)
fw_tools.elf_to_hex(path_elf)



import os
import shutil
from pathlib import Path

import pytest

from shepherd_core import fw_tools

from .conftest import files_elf


@pytest.mark.elf
@pytest.mark.parametrize("path_elf", files_elf)
def test_sym_finding(path_elf: Path) -> None:
    assert fw_tools.find_symbol(path_elf, "SHEPHERD_NODE_ID")


@pytest.mark.elf
@pytest.mark.parametrize("path_elf", files_elf)
def test_sym_reading(path_elf: Path) -> None:
    assert fw_tools.read_symbol(path_elf, "SHEPHERD_NODE_ID", 2)


@pytest.mark.elf
@pytest.mark.parametrize("path_elf", files_elf)
def test_sym_mod(path_elf: Path, tmp_path: Path) -> None:
    value = 0xCAFE
    sym = "SHEPHERD_NODE_ID"
    path_new = tmp_path / path_elf.name
    shutil.copy(path_elf, path_new)
    path_gen = fw_tools.modify_symbol_value(path_new, sym, value, overwrite=False)
    assert path_gen.is_file()
    assert path_gen.as_posix() != path_new.as_posix()
    value_new = fw_tools.read_symbol(path_gen, sym, 2)
    assert value == value_new
    value_old = fw_tools.read_symbol(path_elf, sym, 2)
    assert value_new != value_old


@pytest.mark.elf
@pytest.mark.parametrize("path_elf", files_elf)
def test_sym_mod_overwrite(path_elf: Path, tmp_path: Path) -> None:
    value = 0xCAFE
    sym = "SHEPHERD_NODE_ID"
    assert tmp_path.exists()
    path_new = tmp_path / path_elf.name
    shutil.copy(path_elf, path_new)
    path_gen = fw_tools.modify_symbol_value(path_new, sym, value, overwrite=True)
    if path_gen is None and os.name == "nt":
        # TODO: overwriting fails ATM on WinOS (at least in tmp-dir)
        return
    assert path_gen.is_file()
    assert path_gen.as_posix() == path_new.as_posix()
    value_new = fw_tools.read_symbol(path_gen, sym, 2)
    assert value == value_new
    value_old = fw_tools.read_symbol(path_elf, sym, 2)
    assert value_new != value_old


@pytest.mark.elf
@pytest.mark.parametrize("path_elf", files_elf)
def test_id_mod(path_elf: Path, tmp_path: Path) -> None:
    value = 0xCAFE
    path_new = tmp_path / path_elf.name
    shutil.copy(path_elf, path_new)
    path_gen = fw_tools.modify_uid(path_new, value)
    if path_gen is None and os.name == "nt":
        # TODO: overwriting fails ATM on WinOS (at least in tmp-dir)
        return
    assert path_gen.as_posix() == path_new.as_posix()
    value_new = fw_tools.read_symbol(path_gen, "SHEPHERD_NODE_ID", 2)
    assert value == value_new



"""Read and modify symbols in ELF-files."""

import shutil
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Annotated

from pydantic import Field
from pydantic import validate_call

from shepherd_core.config import config
from shepherd_core.logger import log

from .validation import is_elf

try:
    from pwnlib.elf import ELF
except ImportError as e:
    ELF = None
    elf_error_text = (
        "Please install functionality with 'pip install shepherd-core[elf] -U' first, "
        f"underlying exception: {e.msg}"
    )


@validate_call
def find_symbol(file_elf: Path, symbol: str) -> bool:
    """Find a symbol in the ELF file."""
    if symbol is None or not is_elf(file_elf):
        return False
    if ELF is None:
        raise RuntimeError(elf_error_text)
    with TemporaryDirectory(ignore_cleanup_errors=True) as tmp:
        # switcheroo that might prevent windows bug - overwrite fails in modify_symbol_value()
        file_tmp = Path(tmp) / file_elf.name
        shutil.copy(file_elf, file_tmp)
        elf = ELF(path=file_tmp)
        try:
            addr = elf.symbols[symbol]
        except KeyError:
            addr = None
        if addr is None:
            elf.close()  # better be safe
            log.debug("Symbol '%s' not found in ELF-File %s", symbol, file_elf.name)
            return False
        log.debug(
            "Symbol '%s' found in ELF-File %s, arch=%s, order=%s",
            symbol,
            file_elf.name,
            elf.arch,
            elf.endian,
        )
        elf.close()
    return True


@validate_call
def read_symbol(file_elf: Path, symbol: str, length: int) -> int | None:
    """Read value of symbol in ELF-File.

    Will be interpreted as int.
    """
    if not find_symbol(file_elf, symbol):
        return None
    if ELF is None:
        raise RuntimeError(elf_error_text)
    elf = ELF(path=file_elf)
    addr = elf.symbols[symbol]
    value_raw = elf.read(address=addr, count=length)[-length:]
    endian = elf.endian
    elf.close()
    return int.from_bytes(bytes=value_raw, byteorder=endian, signed=False)


def read_uid(file_elf: Path) -> int | None:
    """Read value of UID-symbol for shepherd testbed."""
    return read_symbol(file_elf, symbol=config.UID_NAME, length=config.UID_SIZE)


def read_arch(file_elf: Path) -> str | None:
    """Determine chip-architecture from elf-metadata."""
    if not is_elf(file_elf):
        return None
    if ELF is None:
        raise RuntimeError(elf_error_text)
    elf = ELF(path=file_elf)
    elf_type = elf.elftype.lower()
    elf_arch = elf.arch.lower()
    elf.close()
    if "exec" in elf_type:
        return elf_arch
    log.error("ELF is not Executable")
    return None


@validate_call
def modify_symbol_value(
    file_elf: Path,
    symbol: str,
    value: Annotated[int, Field(ge=0, lt=2 ** (8 * config.UID_SIZE))],
    *,
    overwrite: bool = False,
) -> Path | None:
    """Replace value of uint16-symbol in ELF-File.

    Hardcoded for uint16_t (2 byte).
    The testbed uses this to patch firmware with custom target-ID.

    NOTE: can overwrite provided file.

    """
    if not find_symbol(file_elf, symbol):
        return None
    if ELF is None:
        raise RuntimeError(elf_error_text)
    with TemporaryDirectory(ignore_cleanup_errors=True) as tmp:
        # switcheroo that also prevents windows bug (overwrite fails)
        file_tmp = Path(tmp) / file_elf.name
        shutil.copy(file_elf, file_tmp)

        elf = ELF(path=file_elf)
        addr = elf.symbols[symbol]
        value_raw = elf.read(address=addr, count=config.UID_SIZE)[-config.UID_SIZE :]
        # ⤷ cutting needed -> msp produces 4b instead of 2
        value_old = int.from_bytes(bytes=value_raw, byteorder=elf.endian, signed=False)
        value_raw = value.to_bytes(length=config.UID_SIZE, byteorder=elf.endian, signed=False)

        try:
            elf.write(address=addr, data=value_raw)
        except AttributeError:
            log.warning("ELF-Modifier failed @%s for symbol '%s'", f"0x{addr:X}", symbol)
            elf.close()
            return None

        file_new = file_elf if overwrite else file_elf.with_stem(file_elf.stem + "_" + str(value))
        try:
            file_new.unlink(missing_ok=True)
        except PermissionError:
            elf.close()
            log.error(
                "Failed to overwrite file, because it's somehow still in use (typical for WinOS)."
            )
            return None
        elf.save(path=file_new)
        elf.close()

    log.debug(
        "Value of Symbol '%s' modified: %s -> %s @%s",
        symbol,
        hex(value_old),
        hex(value),
        hex(addr),
    )
    return file_new


def modify_uid(file_elf: Path, value: int) -> Path | None:
    """Replace value of UID-symbol for shepherd testbed."""
    return modify_symbol_value(file_elf, symbol=config.UID_NAME, value=value, overwrite=True)



"""Config for Task that adds the custom ID to the firmware & stores it into a file."""

from collections.abc import Set as AbstractSet
from pathlib import Path
from pathlib import PurePosixPath
from typing import Annotated
from typing import TypedDict

from pydantic import Field
from pydantic import model_validator
from pydantic import validate_call
from typing_extensions import Self
from typing_extensions import Unpack

from shepherd_core.data_models.base.content import IdInt
from shepherd_core.data_models.base.shepherd import ShpModel
from shepherd_core.data_models.content.firmware import Firmware
from shepherd_core.data_models.content.firmware import FirmwareDType
from shepherd_core.data_models.content.firmware import FirmwareStr
from shepherd_core.data_models.experiment.experiment import Experiment
from shepherd_core.data_models.testbed import Testbed
from shepherd_core.data_models.testbed.target import IdInt16
from shepherd_core.data_models.testbed.target import MCUPort
from shepherd_core.logger import log

from .helper_paths import path_posix


class FirmwareModTask(ShpModel):
    """Config for Task that adds the custom ID to the firmware & stores it into a file."""

    data: FirmwareStr | Path
    data_type: FirmwareDType
    custom_id: IdInt16 | None = None
    firmware_file: Path

    verbose: Annotated[int, Field(ge=0, le=4)] = 2
    """ ⤷ 0=Errors, 1=Warnings, 2=Info, 3=Debug"""

    @model_validator(mode="after")
    def post_validation(self) -> Self:
        if self.data_type in {
            FirmwareDType.base64_hex,
            FirmwareDType.path_hex,
        }:
            log.warning("Firmware is scheduled to get custom-ID but is not in elf-format")
        return self

    @classmethod
    @validate_call
    def from_xp(
        cls,
        xp: Experiment,
        tb: Testbed,
        tgt_id: IdInt,
        mcu_port: MCUPort,
        fw_path: Path,
    ) -> Self:
        tgt_cfg = xp.get_target_config(tgt_id)

        fw = tgt_cfg.firmware1 if mcu_port == 1 else tgt_cfg.firmware2
        if fw is None:
            # TODO: if target has default fw -> use that! otherwise no sleep is flashed
            return None

        fw_id = tgt_cfg.get_custom_id(tgt_id)
        if fw_id is None:
            obs = tb.get_observer(tgt_id)
            fw_id = obs.get_target(tgt_id).testbed_id

        return cls(
            data=path_posix(fw.data) if isinstance(fw.data, Path) else fw.data,
            data_type=fw.data_type,
            custom_id=fw_id,
            firmware_file=path_posix(fw_path),
        )

    @classmethod
    def from_firmware(
        cls,
        fw: Firmware,
        **kwargs: Unpack[TypedDict],
    ) -> Self:
        if not isinstance(fw, Firmware):
            raise TypeError("fw-argument must be of type Firmware")
        kwargs["data"] = fw.data
        kwargs["data_type"] = fw.data_type
        fw.compare_hash()
        path = kwargs.get("firmware_file")
        if path is not None and path.is_dir():
            path_new: Path = path / fw.name
            kwargs["firmware_file"] = path_new.with_suffix(".hex")
        return cls(**kwargs)

    def is_contained(self, paths: AbstractSet[PurePosixPath]) -> bool:
        """Limit paths to allowed directories."""
        all_ok = any(self.firmware_file.is_relative_to(path) for path in paths)
        if isinstance(self.data, Path):
            all_ok = any(self.data.is_relative_to(path) for path in paths)
        return all_ok



"""Client to access a testbed-instance for controlling experiments."""

from .client_abc_fix import tb_client
from .client_web import WebClient
from .user_model import User

__all__ = [
    "User",
    "WebClient",
    "tb_client",
]



"""Generalized path for the file-cache."""

import os
from pathlib import Path


def _get_xdg_path(variable_name: str, default_path: Path) -> Path:
    _value = os.environ.get(variable_name)
    if _value is None or _value == "":
        return default_path
    return Path(_value)


user_path = Path("~").expanduser()

cache_xdg_path = _get_xdg_path("XDG_CACHE_HOME", user_path / ".cache")
cache_user_path = cache_xdg_path / "shepherd"



from pathlib import Path

import pytest
from shepherd_core.data_models import EnergyDType
from shepherd_core.data_models import VirtualHarvesterConfig
from shepherd_core.data_models.content.virtual_harvester_config import HarvesterPRUConfig
from shepherd_core.vsource import VirtualHarvesterModel

from shepherd_core import Reader

hrv_list = [
    "ivcurve",
    "iv1000",
    "isc_voc",
    "cv20",
    "mppt_voc",
    "mppt_bq",
    "mppt_bq_solar",
    "mppt_po",
    "mppt_opt",
]


@pytest.mark.parametrize("hrv_name", hrv_list)
def test_vsource_hrv_min(hrv_name: str) -> None:
    hrv_config = VirtualHarvesterConfig(name=hrv_name)
    hrv_pru = HarvesterPRUConfig.from_vhrv(hrv_config)
    _ = VirtualHarvesterModel(hrv_pru)


def test_vsource_hrv_create_files(
    file_ivcurve: Path, file_ivsample: Path, file_isc_voc: Path
) -> None:
    pass


@pytest.mark.parametrize("hrv_name", hrv_list[:3])
def test_vsource_hrv_fail_ivcurve(hrv_name: str) -> None:
    # the first algos are not usable for ivcurve
    hrv_config = VirtualHarvesterConfig(name=hrv_name)
    with pytest.raises(ValueError):  # noqa: PT011
        _ = HarvesterPRUConfig.from_vhrv(
            hrv_config,
            for_emu=True,
            dtype_in=EnergyDType.ivcurve,
            window_size=100,
            voltage_step_V=0.1,
        )


@pytest.mark.parametrize("hrv_name", hrv_list[3:])
def test_vsource_hrv_sim(hrv_name: str, file_ivcurve: Path) -> None:
    with Reader(file_ivcurve) as file:
        hrv_config = VirtualHarvesterConfig(name=hrv_name)
        hrv_pru = HarvesterPRUConfig.from_vhrv(
            hrv_config,
            for_emu=True,
            dtype_in=file.get_datatype(),
            window_size=file.get_window_samples(),
            voltage_step_V=file.get_voltage_step(),
        )
        hrv = VirtualHarvesterModel(hrv_pru)
        for _t, _v, _i in file.read():
            length = max(_v.size, _i.size)
            for _n in range(length):
                hrv.ivcurve_sample(_voltage_uV=_v[_n] * 10**6, _current_nA=_i[_n] * 10**9)


@pytest.mark.parametrize("hrv_name", hrv_list[3:])
def test_vsource_hrv_fail_isc_voc(hrv_name: str) -> None:
    # not implemented ATM
    hrv_config = VirtualHarvesterConfig(name=hrv_name)
    with pytest.raises(NotImplementedError):
        _ = HarvesterPRUConfig.from_vhrv(hrv_config, for_emu=True, dtype_in=EnergyDType.isc_voc)


def test_vsource_hrv_fail_unknown_type() -> None:
    hrv_config = VirtualHarvesterConfig(name="mppt_voc")
    with pytest.raises(KeyError):
        _ = HarvesterPRUConfig.from_vhrv(hrv_config, for_emu=True, dtype_in="xyz")


def test_vsource_hrv_adapt_voltage_step() -> None:
    hrv_config = VirtualHarvesterConfig(name="mppt_bq")
    # first case without setting step-size
    pru_config1 = HarvesterPRUConfig.from_vhrv(
        hrv_config,
        for_emu=True,
        dtype_in=EnergyDType.ivcurve,
        window_size=1000,
        voltage_step_V=1e-3 * hrv_config.voltage_step_mV,
    )
    step_expected_uV = (pru_config1.voltage_max_uV - pru_config1.voltage_min_uV) / (1000 - 1)
    assert pru_config1.voltage_step_uV > 2 * step_expected_uV
    # now it gets set
    pru_config2 = HarvesterPRUConfig.from_vhrv(
        hrv_config,
        for_emu=True,
        dtype_in=EnergyDType.ivcurve,
        window_size=1000,
        voltage_step_V=1e-6 * step_expected_uV,
    )
    assert pru_config2.voltage_step_uV < 2 * step_expected_uV



import pytest

from shepherd_core import get_verbose_level
from shepherd_core import increase_verbose_level


@pytest.mark.parametrize("log_level", range(-5, 10))
def test_log_levels(log_level: int) -> None:
    increase_verbose_level(log_level)
    if log_level <= 2:
        assert get_verbose_level() == 2
    elif log_level <= 3:
        assert get_verbose_level() == log_level
    else:
        assert get_verbose_level() == 3



"""Config for the Observer in Harvest-Mode to record IV data from a harvesting-source."""

from collections.abc import Set as AbstractSet
from datetime import datetime
from datetime import timedelta
from pathlib import Path
from pathlib import PurePosixPath
from typing import Annotated

from pydantic import Field
from pydantic import model_validator
from typing_extensions import Self

from shepherd_core.data_models.base.shepherd import ShpModel
from shepherd_core.data_models.base.timezone import local_tz
from shepherd_core.data_models.content.virtual_harvester_config import VirtualHarvesterConfig
from shepherd_core.data_models.experiment.observer_features import PowerTracing
from shepherd_core.data_models.experiment.observer_features import SystemLogging

from .emulation import Compression


class HarvestTask(ShpModel):
    """Config for the Observer in Harvest-Mode to record IV data from a harvesting-source."""

    # General config
    output_path: Path
    """ ⤷ dir- or file-path for storing the recorded data:

    - providing a directory -> file is named hrv_timestamp.h5
    - for a complete path the filename is not changed except it exists and
      overwrite is disabled -> name#num.h5
    """
    force_overwrite: bool = False
    """ ⤷ Overwrite existing file"""
    output_compression: Compression | None = Compression.default
    """ ⤷ should be 1 (level 1 gzip), lzf, or None (order of recommendation)"""

    time_start: datetime | None = None
    """ timestamp or unix epoch time, None = ASAP"""
    duration: timedelta | None = None
    """ ⤷ Duration of recording in seconds, None = till EOFSys"""

    # emulation-specific
    use_cal_default: bool = False
    """ ⤷ Use default calibration values, skip loading from EEPROM"""

    virtual_harvester: VirtualHarvesterConfig = VirtualHarvesterConfig(name="mppt_opt")
    """ ⤷ Choose one of the predefined virtual harvesters or configure a new one
    """
    power_tracing: PowerTracing = PowerTracing()
    sys_logging: SystemLogging | None = SystemLogging()

    verbose: Annotated[int, Field(ge=0, le=4)] = 2
    """ ⤷ 0=Errors, 1=Warnings, 2=Info, 3=Debug"""

    # TODO: there is an unused DAC-Output patched to the harvesting-port

    @model_validator(mode="before")
    @classmethod
    def pre_correction(cls, values: dict) -> dict:
        # convert & add local timezone-data, TODO: used twice, refactor
        has_time = values.get("time_start") is not None
        if has_time and isinstance(values["time_start"], (int, float)):
            values["time_start"] = datetime.fromtimestamp(values["time_start"], tz=local_tz())
        if has_time and isinstance(values["time_start"], str):
            values["time_start"] = datetime.fromisoformat(values["time_start"])
        if has_time and values["time_start"].tzinfo is None:
            values["time_start"] = values["time_start"].astimezone()
        return values

    @model_validator(mode="after")
    def post_validation(self) -> Self:
        # TODO: limit paths
        has_time = False  # TODO: deactivated, self.time_start is not None
        time_now = datetime.now().astimezone()
        if has_time and self.time_start < time_now:
            msg = (
                f"Start-Time for Harvest can't be in the past ('{self.time_start}' vs '{time_now}'."
            )
            raise ValueError(msg)
        if self.duration and self.duration.total_seconds() < 0:
            raise ValueError("Task-Duration can't be negative.")
        return self

    def is_contained(self, paths: AbstractSet[PurePosixPath]) -> bool:
        """Limit paths to allowed directories."""
        return any(self.output_path.is_relative_to(path) for path in paths)



"""Base-Model for all content."""

from datetime import datetime
from typing import Annotated
from uuid import uuid4

from pydantic import Field
from pydantic import StringConstraints
from pydantic import model_validator
from typing_extensions import Self

from .shepherd import ShpModel

# constr -> to_lower=True, max_length=16, regex=r"^[\w]+$"
# ⤷ Regex = AlphaNum
IdInt = Annotated[int, Field(ge=0, lt=2**128)]
NameStr = Annotated[str, StringConstraints(max_length=32, pattern=r"^[^<>:;,?\"\*|\/\\]+$")]
# ⤷ Regex = FileSystem-Compatible ASCII
SafeStr = Annotated[str, StringConstraints(pattern=r"^[ -~]+$")]
# ⤷ Regex = All Printable ASCII-Characters with Space


def id_default() -> int:
    """Generate a unique ID - usable as default value.

    Note: IdInt in mongoDB can currently handle 8 bytes (16 hex-values)
    """
    return int(uuid4().hex[-15:], 16)


class ContentModel(ShpModel):
    """Base-Model for content with generalized properties."""

    id: int = Field(
        description="Unique ID",
        default_factory=id_default,
    )
    name: NameStr
    """ ⤷ name of virtual content, models can be queried by this slug."""
    description: Annotated[SafeStr | None, Field(description="Required when public")] = None
    comment: SafeStr | None = None
    created: datetime = Field(default_factory=datetime.now)
    updated_last: datetime = Field(default_factory=datetime.now)
    # TODO: add dedicated 'inherit_from' field?

    # Ownership & Access
    # TODO: remove owner & group, only needed for DB
    owner: NameStr
    group: Annotated[NameStr, Field(description="University or Subgroup")]
    visible2group: bool = False
    visible2all: bool = False

    # TODO: we probably need to remember the lib-version for content &| experiment

    def __str__(self) -> str:
        return self.name

    @model_validator(mode="after")
    def content_validation(self) -> Self:
        is_visible = self.visible2group or self.visible2all
        if is_visible and self.description is None:
            raise ValueError(
                "Public instances require a description (check visible2*- and description-field)"
            )
        return self



"""Helper-functions for firmware-validation.

TODO: Work in Progress.
    - Each arch should have its own file and
    - detection-functions that register in main validator.
"""

import shutil
import tempfile
from pathlib import Path

from intelhex import IntelHex
from intelhex import IntelHexError
from pydantic import validate_call

from shepherd_core.data_models.content.firmware_datatype import FirmwareDType
from shepherd_core.logger import log

from .converter_elf import elf_to_hex

try:
    from elftools.common.exceptions import ELFError
    from pwnlib.elf import ELF
except ImportError as e:
    ELF = None
    ELFError = None
    elf_error_text = (
        "Please install functionality with 'pip install shepherd-core[elf] -U' first, "
        f"underlying exception: {e.msg}"
    )


@validate_call
def is_hex(file: Path) -> bool:
    """Check if file is containing intel-hex data."""
    try:
        _ = IntelHex(file.as_posix())
    except ValueError:  # parsing
        return False
    except IntelHexError:  # structural errors
        return False
    return True


def is_hex_msp430(file: Path) -> bool:
    """Try to detect specifics for that MCU.

    Observations:
    - addresses begin at 0x4000
    - value @0xFFFE (IVT) is start_address (of pgm-code)
    """
    if is_hex(file):
        ih = IntelHex(file.as_posix())
        if ih.minaddr() != 0x4000:
            return False
        if 0xFFFE not in ih.addresses():
            return False
        value = int.from_bytes(ih.gets(0xFFFE, 2), byteorder="little", signed=False)
        if 0x4000 > value >= 0xFF80:
            return False

        # conservative test for now - should be well below 128 kB + 8kB for msp430fr5962
        return ih.get_memory_size() <= 270_000
    return False


def is_hex_nrf52(file: Path) -> bool:
    """Try to detect specifics for that MCU.

    Observations:
    - addresses begin mostly at 0x0 (no must)
    - only one segment (.get_segments), todo
    """
    if is_hex(file):
        ih = IntelHex(file.as_posix())
        # conservative test for now - should be between 0 and 1 MB + 256 kB,
        # but lately showed much higher values > 4 MB
        return ih.get_memory_size() > 0
    return False


# TODO: elf-workflow needs work -> construct experiments without external dependencies
#  - remove conversion to hex
#  - use elftools to verify magic-bytes and similar things done for the hex
#  https://github.com/eliben/pyelftools/wiki/User's-guide


@validate_call
def is_elf(file: Path) -> bool:
    """Check if file is an ELF file."""
    if ELF is None:
        raise RuntimeError(elf_error_text)
    if not file.is_file():
        return False
    try:
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmp:
            # switcheroo that might prevent windows bug - overwrite fails in modify_symbol_value()
            file_tmp = Path(tmp) / file.name
            shutil.copy(file, file_tmp)
            elf = ELF(path=file_tmp)
            elf.close()
    except ELFError:
        log.debug("File %s is not ELF - Magic number does not match", file.name)
        return False
    return True


def is_elf_msp430(file: Path) -> bool:
    """Check if file is an ELF for that MCU.

    TODO: allow detection without conversion
    """
    if is_elf(file):
        with tempfile.TemporaryDirectory() as path:
            file_hex = Path(path) / "file.hex"
            file_hex = elf_to_hex(file, file_hex)
            if is_hex_msp430(file_hex):
                return True
        return False
    return False


def is_elf_nrf52(file: Path) -> bool:
    """Check if file is an ELF for that MCU.

    TODO: allow detection without conversion
    """
    if is_elf(file):
        with tempfile.TemporaryDirectory() as path:
            file_hex = Path(path) / "file.hex"
            file_hex = elf_to_hex(file, file_hex)
            if is_hex_nrf52(file_hex):
                return True
        return False
    return False


def determine_type(file: Path) -> FirmwareDType:
    """Figure out file-type (hex or elf)."""
    if not file.is_file():
        raise ValueError("Fn needs an existing file as input")
    if is_hex(file):
        return FirmwareDType.path_hex
    if is_elf(file):
        return FirmwareDType.path_elf
    msg = f"Type of file '{file.name}' could not be determined"
    raise ValueError(msg)


def determine_arch(file: Path) -> str:
    """Figure out arch (msp430 or nrf52)."""
    file_t = determine_type(file)
    if file_t == FirmwareDType.path_elf:
        if is_elf_msp430(file):
            return "msp430"
        if is_elf_nrf52(file):
            return "nrf52"
        msg = f"Arch of ELF '{file.name}' could not be determined"
        raise ValueError(msg)
    if file_t == FirmwareDType.path_hex:
        if is_hex_msp430(file):
            return "msp430"
        if is_hex_nrf52(file):
            return "nrf52"
        msg = f"Arch of HEX '{file.name}' could not be determined"
        raise ValueError(msg)
    msg = f"Arch of file '{file.name}' could not be determined"
    raise ValueError(msg)



from pathlib import Path

import pytest
from shepherd_core.decoder_waveform import Uart


@pytest.fixture
def example_path() -> Path:
    return Path(__file__).resolve().parent.parent.parent / "examples"


def test_decode_uart(example_path: Path) -> None:
    uwd = Uart(example_path / "uart_raw2.csv")
    _ = uwd.get_symbols()
    _ = uwd.get_lines()
    _ = uwd.get_text()


def test_decode_speedup(example_path: Path) -> None:
    uwd = Uart(example_path / "uart_raw2.csv")
    _ = uwd.get_symbols(force_redo=True)
    _ = uwd.get_lines(force_redo=True)
    _ = uwd.get_text(force_redo=True)

    # these now have cached the result
    _ = uwd.get_symbols()
    _ = uwd.get_lines()
    _ = uwd.get_text()


def test_decode_chained(example_path: Path) -> None:
    uwd = Uart(example_path / "uart_raw2.csv")
    _ = uwd.get_text()  # get_symbols() and get_lines() is executed automatically



"""This example shows ways to embed firmware into the data-model.

Note: the new semi-automatic way to generate a data-model needs pwntools installed
or shepherd-core[elf].
"""

from pathlib import Path

from shepherd_core.data_models import Firmware
from shepherd_core.data_models import FirmwareDType

from shepherd_core import WebClient
from shepherd_core import fw_tools

path_elf = Path(__file__).parent.parent / "tests/fw_tools/build_msp.elf"

# Option 1 - fully manual

fw1 = Firmware(
    name="msp_deep_sleep1",
    data=fw_tools.file_to_base64(path_elf),
    data_type=FirmwareDType.base64_elf,
    mcu={"name": "MSP430FR"},
    owner="example",
    group="test",
)
fw1.to_file(Path(__file__).parent / "firmware.yaml")

# Option 2 - semi-automatic -> MCU and data-type get derived

fw2 = Firmware.from_firmware(
    file=path_elf,
    name="msp_deep_sleep2",
    owner="example",
    group="test",
)

# store embedded data
path_elf2 = fw2.extract_firmware(Path(__file__).parent)
print(f"stored firmware to '{path_elf2.name}'")


# Option 3 - fully automatic (with login) -> owner and group get prefilled

do_connect = False
if do_connect:
    WebClient(token="your_personal_login_token")  # noqa: S106
    fw3 = Firmware.from_firmware(file=path_elf, name="msp_deep_sleep3")



from shepherd_core.data_models import EnergyDType
from shepherd_core.data_models.content.energy_environment import EnergyEnvironment
from shepherd_core.data_models.content.firmware import Firmware
from shepherd_core.data_models.content.virtual_harvester_config import HarvesterPRUConfig
from shepherd_core.data_models.content.virtual_harvester_config import VirtualHarvesterConfig
from shepherd_core.data_models.content.virtual_source_config import ConverterPRUConfig
from shepherd_core.data_models.content.virtual_source_config import VirtualSourceConfig
from shepherd_core.testbed_client.fixtures import Fixtures


def test_testbed_fixture_energy_environment() -> None:
    for fix in Fixtures()["EnergyEnvironment"]:
        EnergyEnvironment(name=fix["name"])
        EnergyEnvironment(id=fix["id"])


def test_testbed_fixture_firmware() -> None:
    for fix in Fixtures()["Firmware"]:
        _id = fix["id"]
        if _id in {1001, 1002}:
            continue
        Firmware(name=fix["name"])
        Firmware(id=fix["id"])


def test_experiment_fixture_vsrc() -> None:
    for fix in Fixtures()["VirtualSourceConfig"]:
        vsrc = VirtualSourceConfig(name=fix["name"])
        VirtualSourceConfig(id=fix["id"])
        ConverterPRUConfig.from_vsrc(
            vsrc, dtype_in=EnergyDType.ivcurve, log_intermediate_node=False
        )
        ConverterPRUConfig.from_vsrc(
            vsrc, dtype_in=EnergyDType.ivsample, log_intermediate_node=False
        )
        ConverterPRUConfig.from_vsrc(
            vsrc, dtype_in=EnergyDType.ivsample, log_intermediate_node=True
        )


def test_experiment_fixture_vhrv() -> None:
    for fix in Fixtures()["VirtualHarvesterConfig"]:
        if fix["name"] == "neutral":
            continue
        vhrv = VirtualHarvesterConfig(name=fix["name"])
        VirtualHarvesterConfig(id=fix["id"])
        if int(fix["id"]) < 3000:
            HarvesterPRUConfig.from_vhrv(vhrv, for_emu=False)
        if int(fix["id"]) >= 2000:
            HarvesterPRUConfig.from_vhrv(vhrv, for_emu=True)



"""Module for content-type data-models.

These models import externally from: /base, /testbed.
"""

from .energy_environment import EnergyDType
from .energy_environment import EnergyEnvironment
from .firmware import Firmware
from .firmware_datatype import FirmwareDType
from .virtual_harvester_config import VirtualHarvesterConfig
from .virtual_source_config import VirtualSourceConfig
from .virtual_storage_config import VirtualStorageConfig

__all__ = [
    "EnergyDType",
    "EnergyEnvironment",
    "Firmware",
    "FirmwareDType",
    "VirtualHarvesterConfig",
    "VirtualSourceConfig",
    "VirtualStorageConfig",
]



"""Shows how to use inventory functionality."""

from pathlib import Path

import shepherd_core.inventory as si

pyt_inv = si.PythonInventory.collect()
print(f"PyInv: {pyt_inv}")
sys_inv = si.SystemInventory.collect()
print(f"SysInv: {sys_inv}")
tgt_inv = si.TargetInventory.collect()
print(f"TgtInv: {tgt_inv}")


inv = si.Inventory.collect()
print(f"Complete Inventory: {inv}")
inv.to_file("inventory.yaml", minimal=True, comment="just a test")

inl = si.InventoryList(elements=[inv])
inl.to_csv(Path(__file__).parent / "inventory.csv")



"""Models for the calibration-data to convert between raw & SI-Values."""

import struct
from collections.abc import Callable
from collections.abc import Generator
from collections.abc import Mapping
from collections.abc import Sequence
from typing import TypeVar

import numpy as np
from numpy.typing import NDArray
from pydantic import Field
from pydantic import PositiveFloat
from pydantic import conbytes
from pydantic import constr
from pydantic import validate_call
from typing_extensions import Self

from shepherd_core.calibration_hw_def import adc_current_to_raw
from shepherd_core.calibration_hw_def import adc_voltage_to_raw
from shepherd_core.calibration_hw_def import dac_voltage_to_raw

from .shepherd import ShpModel
from .timezone import local_iso_date

Calc_t = TypeVar("Calc_t", NDArray[np.float64], float)


def dict_generator(
    in_dict: Mapping | Sequence, pre: list | None = None
) -> Generator[list, None, None]:
    """Recursive helper-function to generate a 1D-List(or not?).

    TODO: isn't that a 1D-List generator?
    """
    pre = pre[:] if pre else []
    if isinstance(in_dict, Mapping):
        for key, value in in_dict.items():
            if isinstance(value, Mapping):
                yield from dict_generator(value, [*pre, key])
            elif isinstance(value, Sequence):
                for v in value:
                    yield from dict_generator(v, [*pre, key])
            else:
                yield [*pre, key, value]
    else:
        yield [*pre, in_dict]


class CalibrationPair(ShpModel):
    """SI-value [SI-Unit] = raw-value * gain + offset."""

    gain: PositiveFloat
    offset: float = 0
    unit: str | None = None  # TODO: add units when used

    def raw_to_si(self, values_raw: Calc_t, *, allow_negative: bool = True) -> Calc_t:
        """Convert between physical units and raw unsigned integers."""
        values_si = values_raw * self.gain + self.offset
        if not allow_negative:
            if isinstance(values_si, np.ndarray):
                values_si[values_si < 0.0] = 0.0
                # if pyright still complains, cast with .astype(float)
            else:
                values_si = float(max(values_si, 0.0))
        elif not isinstance(values_si, np.ndarray):
            values_si = float(values_si)
        return values_si

    def si_to_raw(self, values_si: Calc_t) -> Calc_t:
        """Convert between physical units and raw unsigned integers."""
        values_raw = (values_si - self.offset) / self.gain
        if isinstance(values_raw, np.ndarray):
            values_raw[values_raw < 0.0] = 0.0
            values_raw = np.around(values_raw)
            # TODO: overflow should also be prevented (add bit-width) -> fail or warn at both?
        else:
            values_raw = round(max(values_raw, 0.0))
        return values_raw

    @classmethod
    def from_fn(cls, fn: Callable, unit: str | None = None) -> Self:
        """Probe linear function to determine scaling values."""
        offset = fn(0, limited=False)
        gain_inv = fn(1.0, limited=False) - offset
        return cls(gain=1.0 / float(gain_inv), offset=-float(offset) / gain_inv, unit=unit)


cal_hrv_legacy = {  # legacy translator
    "dac_voltage_a": "dac_V_Sim",
    "dac_voltage_b": "dac_V_Hrv",
    "adc_current": "adc_C_Hrv",  # datalog current
    "adc_voltage": "adc_V_Sense",  # datalog voltage
}

# defaults (pre-init complex types)
cal_pair_dac_V = CalibrationPair.from_fn(dac_voltage_to_raw, unit="V")
cal_pair_adc_V = CalibrationPair.from_fn(adc_voltage_to_raw, unit="V")
cal_pair_adc_C = CalibrationPair.from_fn(adc_current_to_raw, unit="A")


class CalibrationHarvester(ShpModel):
    """Container for all calibration-pairs for that device."""

    dac_V_Hrv: CalibrationPair = cal_pair_dac_V
    dac_V_Sim: CalibrationPair = cal_pair_dac_V
    adc_V_Sense: CalibrationPair = cal_pair_adc_V
    adc_C_Hrv: CalibrationPair = cal_pair_adc_C

    def export_for_sysfs(self) -> dict:
        """Convert and write the essential data.

        [scaling according to commons.h]
        # ADC-C is handled in nA (nano-ampere), gain is shifted by 8 bit
        # ADC-V is handled in uV (micro-volt), gain is shifted by 8 bit
        # DAC-V is handled in uV (micro-volt), gain is shifted by 20 bit
        """
        cal_set = {
            "adc_current_gain": round(1e9 * (2**8) * self.adc_C_Hrv.gain),
            "adc_current_offset": round(1e9 * (2**0) * self.adc_C_Hrv.offset),
            "adc_voltage_gain": round(1e6 * (2**8) * self.adc_V_Sense.gain),
            "adc_voltage_offset": round(1e6 * (2**0) * self.adc_V_Sense.offset),
            "dac_voltage_gain": round((2**20) / (1e6 * self.dac_V_Hrv.gain)),
            "dac_voltage_offset": round(1e6 * (2**0) * self.dac_V_Hrv.offset),
        }
        for key, value in cal_set.items():
            if (("gain" in key) and not (0 <= value < 2**32)) or (
                ("offset" in key) and not (-(2**31) <= value < 2**31)
            ):
                msg = f"Value ({key}={value}) exceeds uint32-container"
                raise ValueError(msg)
        return cal_set


cal_emu_legacy = {  # legacy translator
    "dac_voltage_a": "dac_V_A",
    "dac_voltage_b": "dac_V_B",  # datalog voltage
    "adc_current": "adc_C_A",  # datalog current
    "adc_voltage": "adc_C_B",
}


class CalibrationEmulator(ShpModel):
    """Container for all calibration-pairs for that device.

    Differentiates between both target-ports A/B.
    """

    dac_V_A: CalibrationPair = cal_pair_dac_V
    dac_V_B: CalibrationPair = cal_pair_dac_V
    adc_C_A: CalibrationPair = cal_pair_adc_C
    adc_C_B: CalibrationPair = cal_pair_adc_C

    def export_for_sysfs(self) -> dict:
        """Convert and write the essential data.

        [scaling according to commons.h]
        # ADC-C is handled in nA (nano-ampere), gain is shifted by 8 bit
        # ADC-V -> unused by vsrc / emu
        # DAC-V is handled in uV (micro-volt), gain is shifted by 20 bit
        TODO: current impl does not distinguish target_ports for DAC
        """
        cal_set = {
            "adc_current_gain": round(1e9 * (2**8) * self.adc_C_A.gain),
            "adc_current_offset": round(1e9 * (2**0) * self.adc_C_A.offset),
            "adc_voltage_gain": round(1e6 * (2**8) * 1),
            "adc_voltage_offset": round(1e6 * (2**0) * 0),
            "dac_voltage_gain": round((2**20) / (1e6 * self.dac_V_A.gain)),
            "dac_voltage_offset": round(1e6 * (2**0) * self.dac_V_A.offset),
        }
        for key, value in cal_set.items():
            if (("gain" in key) and not (0 <= value < 2**32)) or (
                ("offset" in key) and not (-(2**31) <= value < 2**31)
            ):
                msg = f"Value ({key}={value}) exceeds uint32-container"
                raise ValueError(msg)
        return cal_set


class CapeData(ShpModel):
    """Representation of Beaglebone Cape information.

    User must at least provide serial-number on creation.

    According to BeagleBone specifications, each cape should host an EEPROM
    that contains some standardized information about the type of cape,
    manufacturer, version etc.

    `See<https://github.com/beagleboard/beaglebone-black/wiki/System-Reference-Manual#824_EEPROM_Data_Format>`_
    """

    header: conbytes(max_length=4) = b"\xaa\x55\x33\xee"
    eeprom_revision: constr(max_length=2) = "A2"
    board_name: constr(max_length=32) = "BeagleBone SHEPHERD2 Cape"
    version: constr(max_length=4) = "24B0"
    manufacturer: constr(max_length=16) = "NES TU DRESDEN"
    part_number: constr(max_length=16) = "BB-SHPRD"

    serial_number: constr(max_length=12)

    cal_date: constr(max_length=12) = Field(default_factory=local_iso_date)
    # ⤷ produces something like '2023-01-01'

    def __repr__(self) -> str:  # TODO: override useful?
        """string-representation allows print(model)."""
        return str(self.model_dump())


class CalibrationCape(ShpModel):
    """Represents calibration data of shepherd cape.

    Defines the format of calibration data and provides convenient functions
    to read and write calibration data.

    YAML: .to_file() and .from_file() already in ShpModel
    """

    cape: CapeData | None = None
    host: str | None = None

    harvester: CalibrationHarvester = CalibrationHarvester()
    emulator: CalibrationEmulator = CalibrationEmulator()

    @classmethod
    def from_bytestr(cls, data: bytes, cape: CapeData | None = None) -> Self:
        """Instantiate calibration data based on byte string.

        This is mainly used to deserialize data read from an EEPROM memory.

        Args:
        ----
            data: Byte string containing calibration data.
            cape: data can be supplied
        Returns:
            CalibrationCape object with extracted calibration data.

        """
        dv = cls().model_dump(include={"harvester", "emulator"})
        lw1 = list(dict_generator(dv))
        lw2 = [elem for elem in lw1 if isinstance(elem[-1], float)]
        values = struct.unpack(">" + len(lw2) * "d", data)
        # ⤷ X => double float, big endian
        for _i, walk in enumerate(lw2):
            # hardcoded fixed depth ... bad but easy
            dv[walk[0]][walk[1]][walk[2]] = float(values[_i])
        dv["cape"] = cape
        return cls(**dv)

    def to_bytestr(self) -> bytes:
        """Serialize calibration data to byte string.

        Used to prepare data for writing it to EEPROM.

        Returns:
            Byte string representation of calibration values.

        """
        lw = list(dict_generator(self.model_dump(include={"harvester", "emulator"})))
        values = [walk[-1] for walk in lw if isinstance(walk[-1], float)]
        return struct.pack(">" + len(values) * "d", *values)


class CalibrationSeries(ShpModel):
    """Cal-Data for a typical recording of a testbed experiment."""

    voltage: CalibrationPair = CalibrationPair(gain=3 * 1e-9, unit="V")
    # ⤷ default allows 0 - 12 V in 3 nV-Steps
    current: CalibrationPair = CalibrationPair(gain=250 * 1e-12, unit="A")
    # ⤷ default allows 0 - 1 A in 250 pA - Steps
    time: CalibrationPair = CalibrationPair(gain=1e-9, unit="s")
    # ⤷ default = nanoseconds

    @classmethod
    @validate_call(validate_return=False)
    def from_cal(
        cls,
        cal: CalibrationHarvester | CalibrationEmulator,
        *,
        emu_port_a: bool = True,
    ) -> Self:
        if isinstance(cal, CalibrationHarvester):
            return cls(voltage=cal.adc_V_Sense, current=cal.adc_C_Hrv)
        if emu_port_a:
            return cls(voltage=cal.dac_V_A, current=cal.adc_C_A)
        return cls(voltage=cal.dac_V_B, current=cal.adc_C_B)



from pathlib import Path

import pytest
from shepherd_core.data_models import FirmwareDType

from shepherd_core import fw_tools

from .conftest import files_elf


@pytest.mark.elf
@pytest.mark.converter
@pytest.mark.parametrize("path_elf", files_elf)
def test_elf_detection(path_elf: Path) -> None:
    assert fw_tools.is_elf(path_elf)
    if "nrf" in path_elf.name:
        assert fw_tools.is_elf_nrf52(path_elf)
    elif "msp" in path_elf.name:
        assert fw_tools.is_elf_msp430(path_elf)


@pytest.mark.elf
@pytest.mark.converter
@pytest.mark.parametrize("path_elf", files_elf)
def test_elf_determination(path_elf: Path) -> None:
    assert fw_tools.determine_type(path_elf) == FirmwareDType.path_elf
    if "nrf" in path_elf.name:
        assert fw_tools.determine_arch(path_elf) == "nrf52"
    elif "msp" in path_elf.name:
        assert fw_tools.determine_arch(path_elf) == "msp430"


@pytest.mark.elf
@pytest.mark.converter
@pytest.mark.parametrize("path_elf", files_elf)
def test_hex_detection(path_elf: Path, tmp_path: Path) -> None:
    path_hex = (tmp_path / (path_elf.stem + ".hex")).resolve()
    path_hex = fw_tools.elf_to_hex(path_elf, path_hex)
    assert fw_tools.is_hex(path_hex)
    if "nrf" in path_elf.name:
        assert fw_tools.is_hex_nrf52(path_hex)
    elif "msp" in path_elf.name:
        assert fw_tools.is_hex_msp430(path_hex)


@pytest.mark.elf
@pytest.mark.converter
@pytest.mark.parametrize("path_elf", files_elf)
def test_hex_determination(path_elf: Path, tmp_path: Path) -> None:
    path_hex = (tmp_path / (path_elf.stem + ".hex")).resolve()
    path_hex = fw_tools.elf_to_hex(path_elf, path_hex)
    assert fw_tools.determine_type(path_hex) == FirmwareDType.path_hex
    if "nrf" in path_elf.name:
        assert fw_tools.determine_arch(path_hex) == "nrf52"
    elif "msp" in path_elf.name:
        assert fw_tools.determine_arch(path_hex) == "msp430"



from pathlib import Path

from shepherd_core.data_models.content import VirtualSourceConfig
from shepherd_core.data_models.experiment import Experiment
from shepherd_core.data_models.task import EmulationTask
from shepherd_core.data_models.task import HarvestTask
from shepherd_core.data_models.testbed.testbed import Testbed as TasteBad

from .conftest import load_yaml

# ⤷ TasteBad avoids pytest-warning


def test_example_emu() -> None:
    data_dict = load_yaml("example_config_emulator.yaml")
    assert data_dict["mode"] == "emulator"
    emu = EmulationTask(**data_dict["parameters"])
    print(emu)


def test_example_hrv() -> None:
    data_dict = load_yaml("example_config_harvester.yaml")
    assert data_dict["mode"] == "harvester"
    emu = HarvestTask(**data_dict["parameters"])
    print(emu)


def test_example_exp_recommended() -> None:
    # new way
    path = Path(__file__).with_name("example_config_experiment.yaml")
    Experiment.from_file(path)


def test_example_exp() -> None:
    # non-optimal / old way
    data_dict = load_yaml("example_config_experiment_alternative.yaml")
    Experiment(**data_dict)


def test_example_tb() -> None:
    data_dict = load_yaml("example_config_testbed.yaml")
    print(data_dict)
    TasteBad(**data_dict)


def test_example_vsrc() -> None:
    data_dict = load_yaml("example_config_virtsource.yaml")
    VirtualSourceConfig(**data_dict["VirtualSource"])



import os
import subprocess
import sys
from pathlib import Path

import pytest

path_example = Path(__file__).resolve().parent.parent / "examples"
path_simulation = Path(__file__).resolve().parent.parent / "examples/simulations"

simulations: list[str] = [
    "vharvester.py",
    "vsource.py",
    "vstorage.py",
]


@pytest.mark.parametrize("file", simulations)
def test_simulation_scripts(file: str) -> None:
    path_file = path_simulation / file
    os.chdir(path_file.parent)
    subprocess.run([sys.executable, path_file.as_posix()], shell=True, check=True)


examples: list[str] = [
    "experiment_generic_var1.py",
    "experiment_models.py",
    "inventory.py",
    "uart_decode_waveform.py",
    "vsource_debug_sim.py",
]


@pytest.mark.parametrize("file", examples)
def test_example_scripts(file: str) -> None:
    path_file = path_example / file
    os.chdir(path_file.parent)
    subprocess.run([sys.executable, path_file.as_posix()], shell=True, check=True)


examples_fw: list[str] = [
    "firmware_model.py",
    "firmware_modification.py",
]


@pytest.mark.converter
@pytest.mark.elf
@pytest.mark.parametrize("file", examples_fw)
def test_example_scripts_fw(file: str) -> None:
    path_file = path_example / file
    os.chdir(path_file.parent)
    subprocess.run([sys.executable, path_file.as_posix()], shell=True, check=True)



"""Shows different ways to define an experiment.

How to define an experiment:

- within python (shown in this example)
    - object-oriented data-models of
        - experiment
        - TargetConfig -> shared for group of targets
        - virtualSource -> defines energy environment and converters
    - sub-elements reusable
    - scriptable for range of experiments
    - check for plausibility right away
- as yaml (shown in experiment_from_yaml.yaml)
    - default file-format for storing meta-data (for shepherd)
    - minimal writing
    - easy to copy parts
    - submittable through web-interface

"""

import shepherd_core.data_models as sm
from shepherd_core.data_models.task import TestbedTasks

from shepherd_core import WebClient

# generate description for all parameters -> base for web-forms
sm.Experiment.schema_to_file("experiment_schema.yaml")

# For online-queries the lib can be connected to the testbed-server.
# NOTE: there are 3 states:
# - unconnected -> demo-fixtures are queried (locally)
# - connected -> publicly available data is queried online
# - logged in with valid token -> also private data is queried online
do_connect = False
if do_connect:
    WebClient()

# Defining an Experiment in Python
hrv = sm.VirtualHarvesterConfig(name="mppt_bq_thermoelectric")

target_cfgs = [
    # first Instance similar to yaml-syntax
    sm.TargetConfig(
        target_IDs=[9, 10, 11],
        custom_IDs=[0, 1, 2],
        energy_env={"name": "SolarSunny"},
        virtual_source={"name": "diode+capacitor"},
        firmware1={"name": "nrf52_demo_rf"},
    ),
    # second Instance fully object-oriented (recommended)
    sm.TargetConfig(
        target_IDs=list(range(1, 5)),
        custom_IDs=list(range(7, 18)),  # note: longer list is OK
        energy_env=sm.EnergyEnvironment(name="ThermoelectricWashingMachine"),
        virtual_source=sm.VirtualSourceConfig(name="BQ25570-Schmitt", harvester=hrv),
        firmware1=sm.Firmware(name="nrf52_demo_rf"),
        firmware2=sm.Firmware(name="msp430_deep_sleep"),
    ),
]

xperi1 = sm.Experiment(
    name="meaningful Test-Name",
    time_start="2033-03-13 14:15:16",  # or: datetime.now() + timedelta(minutes=30)
    target_configs=target_cfgs,
)

# Safe, reload and compare content
xperi1.to_file("experiment_from_py.yaml", minimal=False)
xperi2 = sm.Experiment.from_file("experiment_from_py.yaml")
print(f"xp1 hash: {xperi1.get_hash()}")
print(f"xp2 hash: {xperi2.get_hash()}")

# comparison to same config (in yaml) fails due to internal variables, BUT:
xperi3 = sm.Experiment.from_file("experiment_from_yaml.yaml")
print(f"xp3 hash: {xperi3.get_hash()} (won't match)")

# Create a tasks-list for the testbed
tb_tasks2 = TestbedTasks.from_xp(xperi2)
tb_tasks2.to_file("experiment_tb_tasks.yaml")

# Comparison between task-Lists succeed (experiment-comparison failed)
tb_tasks3 = TestbedTasks.from_xp(xperi3)
print(f"tasks2 hash: {tb_tasks2.get_hash()}")
print(f"tasks3 hash: {tb_tasks3.get_hash()}")



"""meta-data representation of a testbed-component (physical object)."""

from datetime import date
from datetime import datetime
from enum import Enum
from typing import Any

from pydantic import Field
from pydantic import model_validator

from shepherd_core.data_models.base.content import IdInt
from shepherd_core.data_models.base.content import NameStr
from shepherd_core.data_models.base.content import SafeStr
from shepherd_core.data_models.base.shepherd import ShpModel
from shepherd_core.testbed_client import tb_client


class TargetPort(str, Enum):
    """Options for choosing a target-port."""

    A = a = "A"
    B = b = "B"


class Cape(ShpModel, title="Shepherd-Cape"):
    """meta-data representation of a testbed-component (physical object)."""

    id: IdInt
    name: NameStr
    version: NameStr
    description: SafeStr
    comment: SafeStr | None = None
    # TODO: wake_interval, calibration

    active: bool = True
    created: date | datetime = Field(default_factory=datetime.now)
    calibrated: date | datetime | None = None

    def __str__(self) -> str:
        return self.name

    @model_validator(mode="before")
    @classmethod
    def query_database(cls, values: dict[str, Any]) -> dict[str, Any]:
        values, _ = tb_client.try_completing_model(cls.__name__, values)
        return values



"""meta-data representation of a testbed-component (physical object)."""

from enum import Enum
from typing import Annotated
from typing import Any

from pydantic import Field
from pydantic import StringConstraints
from pydantic import model_validator
from typing_extensions import Self

from shepherd_core.data_models.base.content import IdInt
from shepherd_core.data_models.base.content import NameStr
from shepherd_core.data_models.base.content import SafeStr
from shepherd_core.data_models.base.shepherd import ShpModel
from shepherd_core.testbed_client import tb_client


class Direction(str, Enum):
    """Options for pin-direction."""

    Input = IN = "IN"
    Output = OUT = "OUT"
    Bidirectional = IO = "IO"


class GPIO(ShpModel, title="GPIO of Observer Node"):
    """meta-data representation of a testbed-component."""

    id: IdInt
    name: NameStr
    description: SafeStr | None = None
    comment: SafeStr | None = None

    direction: Direction = Direction.Input
    dir_switch: Annotated[str, StringConstraints(max_length=32)] | None = None

    reg_pru: Annotated[str, StringConstraints(max_length=10)] | None = None
    pin_pru: Annotated[str, StringConstraints(max_length=10)] | None = None
    reg_sys: Annotated[int, Field(ge=0)] | None = None
    pin_sys: Annotated[str, StringConstraints(max_length=10)] | None = None

    def __str__(self) -> str:
        return self.name

    @model_validator(mode="before")
    @classmethod
    def query_database(cls, values: dict[str, Any]) -> dict[str, Any]:
        values, _ = tb_client.try_completing_model(cls.__name__, values)
        return values

    @model_validator(mode="after")
    def post_validation(self) -> Self:
        # ensure that either pru or sys is used, otherwise instance is considered faulty
        no_pru = (self.reg_pru is None) or (self.pin_pru is None)
        no_sys = (self.reg_sys is None) or (self.pin_sys is None)
        if no_pru and no_sys:
            msg = (
                "GPIO-Instance is faulty -> "
                f"it needs to use pru or sys, content: {self.model_dump()}"
            )
            raise ValueError(msg)
        return self

    def user_controllable(self) -> bool:
        return ("gpio" in self.name.lower()) and (self.direction in {"IO", "OUT"})

    def user_recordable(self) -> bool:
        return (
            ("gpio" in self.name.lower())
            and (self.direction in {"IO", "IN"})
            and (self.pin_pru is not None)
        )



from pathlib import Path

import pytest
from shepherd_core.inventory import Inventory
from shepherd_core.inventory import InventoryList
from shepherd_core.inventory import PythonInventory
from shepherd_core.inventory import SystemInventory
from shepherd_core.inventory import TargetInventory


@pytest.mark.parametrize("inv", [PythonInventory, SystemInventory, TargetInventory, Inventory])
def test_collect_data(inv: Inventory) -> None:
    inv.collect()


def test_inventorize(tmp_path: Path) -> None:
    inv = Inventory.collect()
    inl = InventoryList(elements=[inv])
    inl.to_csv(tmp_path / "some.csv")



from pathlib import Path

import pytest
from pydantic import ValidationError
from shepherd_core.data_models import Experiment
from shepherd_core.data_models import FirmwareDType
from shepherd_core.data_models import GpioActuation
from shepherd_core.data_models import GpioEvent
from shepherd_core.data_models.task import ObserverTasks
from shepherd_core.data_models.task.emulation import EmulationTask
from shepherd_core.data_models.task.firmware_mod import FirmwareModTask
from shepherd_core.data_models.task.harvest import HarvestTask
from shepherd_core.data_models.task.programming import ProgrammingTask
from shepherd_core.data_models.testbed import GPIO
from shepherd_core.data_models.testbed import ProgrammerProtocol
from shepherd_core.data_models.testbed import Testbed as TasteBad


def test_task_model_emu_min() -> None:
    EmulationTask(
        input_path="./here",
    )


@pytest.mark.skip(reason="Relaxed in v2025.08.01")
def test_task_model_emu_fault_in_past() -> None:
    with pytest.raises(ValidationError):
        EmulationTask(
            input_path="./here",
            time_start="1984-01-01 11:12:13",
        )


@pytest.mark.parametrize("value", [0, 1, 2, 3, 4, 4.5, "buffer", "main"])
def test_task_model_emu_custom_aux(value: float | str) -> None:
    EmulationTask(
        input_path="./here",
        voltage_aux=value,
    )


@pytest.mark.parametrize("value", [-1.0, 5, "max", "something"])
def test_task_model_emu_fault_aux(value: float | str) -> None:
    with pytest.raises(ValidationError):
        EmulationTask(
            input_path="./here",
            voltage_aux=value,
        )


def test_task_model_emu_fault_gpio_actuation() -> None:
    with pytest.raises(ValidationError):
        EmulationTask(
            input_path="./here",
            gpio_actuation=GpioActuation(
                events=[GpioEvent(delay=5, gpio=GPIO(name="GPIO5"))],
            ),
        )


def test_task_model_fw_min() -> None:
    FirmwareModTask(
        data=Path("/"),
        data_type=FirmwareDType.path_elf,
        custom_id=42,
        firmware_file=Path("fw_to_be.elf"),
    )


def test_task_model_fw_fault_hex() -> None:
    # just a warning for now
    FirmwareModTask(
        data=Path("/"),
        data_type=FirmwareDType.path_hex,
        custom_id=42,
        firmware_file=Path("fw_to_be.hex"),
    )


def test_task_model_hrv_min() -> None:
    HarvestTask(
        output_path="./here",
    )


def test_task_model_hrv_duration() -> None:
    hrv = HarvestTask(
        output_path="./here",
        duration=42,
    )
    assert hrv.duration.total_seconds() == 42


@pytest.mark.skip(reason="Relaxed in v2025.08.01")
def test_task_model_hrv_too_late() -> None:
    with pytest.raises(ValidationError):
        HarvestTask(
            output_path="./here",
            time_start="1984-01-01 11:12:13",
        )


def test_task_model_observer_min1() -> None:
    ObserverTasks(
        observer="peeping tom",
        time_prep="2044-01-01 12:13:14",
        root_path="/usr",
    )


def test_task_model_observer_min2() -> None:
    path = Path(__file__).with_name("example_config_experiment.yaml")
    xp = Experiment.from_file(path)
    ObserverTasks.from_xp(xp=xp, xp_folder=None, tb=TasteBad(name="shepherd_tud_nes"), tgt_id=1)


def test_task_model_prog_min() -> None:
    ProgrammingTask(
        firmware_file=Path("fw_to_load.hex"),
        protocol=ProgrammerProtocol.SWD,
        mcu_type="nrf52",
    )


def test_task_model_prog_fault_elf() -> None:
    with pytest.raises(ValidationError):
        ProgrammingTask(
            firmware_file=Path("fw_to_load.elf"),
            protocol=ProgrammerProtocol.SWD,
            mcu_type="nrf52",
        )



"""Module for experiment related data-models.

These models import externally from: /base, /content, /testbed.
"""

from .experiment import Experiment
from .observer_features import GpioActuation
from .observer_features import GpioEvent
from .observer_features import GpioLevel
from .observer_features import GpioTracing
from .observer_features import PowerTracing
from .observer_features import SystemLogging
from .target_config import TargetConfig

__all__ = [
    "Experiment",
    "GpioActuation",
    "GpioEvent",
    "GpioLevel",
    "GpioTracing",
    "PowerTracing",
    "SystemLogging",
    "TargetConfig",
]



"""A set of experiments to validate and qualify the virtual storage algorithms.

Some general Notes:

- ShpCap is also displayed when emulating Lipo and lead-acid, but it can't and won't behave similar
- during charging the model KiBaM-Plus will deviate from normal KiBaM and KiBaM-Simple,
  as it supports the rate capacity effect and transients (during charging)

"""

import multiprocessing
import os
import sys
from datetime import timedelta
from pathlib import Path

from pydantic import PositiveFloat
from pydantic import validate_call
from shepherd_core.data_models.content.virtual_storage_config import VirtualStorageConfig
from shepherd_core.data_models.content.virtual_storage_config import soc_t
from shepherd_core.vsource.virtual_storage_model import ModelStorage
from shepherd_core.vsource.virtual_storage_model import VirtualStorageModel
from shepherd_core.vsource.virtual_storage_models_kibam import ModelKiBaM
from shepherd_core.vsource.virtual_storage_models_kibam import ModelKiBaMPlus
from shepherd_core.vsource.virtual_storage_models_kibam import ModelKiBaMSimple
from shepherd_core.vsource.virtual_storage_models_kibam import ModelShpCap
from shepherd_core.vsource.virtual_storage_simulator import StorageSimulator

from shepherd_core import log

path_here = Path(__file__).parent

duration_max = 20 if "PYTEST_CURRENT_TEST" in os.environ else sys.float_info.max
# ⤷ limits runtime for pytest


@validate_call
def get_models(
    SoC_init: soc_t, config: VirtualStorageConfig, dt_s: PositiveFloat
) -> list[ModelStorage]:
    """Models to include in experiments."""
    return [
        ModelKiBaM(SoC_init=SoC_init, cfg=config, dt_s=dt_s),
        ModelKiBaMPlus(SoC_init=SoC_init, cfg=config, dt_s=dt_s),
        ModelKiBaMSimple(SoC_init=SoC_init, cfg=config, dt_s=dt_s, optimize_clamp=True),
        ModelKiBaMSimple(SoC_init=SoC_init, cfg=config, dt_s=dt_s, interpolate=True),
        VirtualStorageModel(SoC_init=SoC_init, cfg=config, dt_s=dt_s),
        ModelShpCap(SoC_init=SoC_init, cfg=config, dt_s=dt_s),
    ][1:5]


class CurrentPulsed:
    """A simple constant current source that is pulsed until a target SoC is reached."""

    @validate_call
    def __init__(
        self,
        I_pulse: float,
        period_pulse: PositiveFloat,
        duration_pulse: PositiveFloat,
        SoC_target: soc_t,
    ) -> None:
        self.I_pulse = I_pulse
        self.period_pulse = period_pulse
        self.duration_pulse = duration_pulse
        self.SoC_target = SoC_target

    def step(self, t_s: float, SoC: float, _v: float) -> float:
        if (self.I_pulse < 0 and SoC <= self.SoC_target) or (
            self.I_pulse > 0 and SoC >= self.SoC_target
        ):
            return 0
        return self.I_pulse if t_s % self.period_pulse < self.duration_pulse else 0


class ResistiveChargePulsed:
    """A pulsed charger that is 'current limited' by a resistor."""

    @validate_call
    def __init__(
        self,
        V_target: PositiveFloat,
        R_Ohm: PositiveFloat,
        period_pulse: PositiveFloat,
        duration_pulse: PositiveFloat,
    ) -> None:
        self.R_Ohm = R_Ohm
        self.V_target = V_target
        self.period_pulse = period_pulse
        self.duration_pulse = duration_pulse

    def step(self, t_s: float, _s: float, V: float) -> float:
        I_A = (self.V_target - V) / self.R_Ohm
        return I_A if t_s % self.period_pulse < self.duration_pulse else 0


def experiment_current_ramp_pos(config: VirtualStorageConfig) -> None:
    """Charge virtual storage with a positive current ramp (increasing power)."""
    dt_s = 0.1
    SoC_start = 0.5
    duration_s = min(200, duration_max)
    sim = StorageSimulator(
        models=get_models(SoC_start, config, dt_s),
        dt_s=dt_s,
    )

    def current_trace(t_s: float, _s: float, _v: float) -> float:
        return 0.1 + 0.15 * t_s / duration_s  # pru-model can handle +- 268 mA

    sim.run(fn=current_trace, duration_s=duration_s)
    sim.plot(path_here, f"XP {config.name}, current charge ramp (positive)")


def experiment_current_ramp_neg(config: VirtualStorageConfig) -> None:
    """Discharge virtual storage with a negative current ramp (increasing power)."""
    dt_s = 0.1
    SoC_start = 0.5
    duration_s = min(200, duration_max)
    sim = StorageSimulator(
        models=get_models(SoC_start, config, dt_s),
        dt_s=dt_s,
    )

    def current_trace(t_s: float, _s: float, _v: float) -> float:
        return -(0.1 + 0.14 * t_s / duration_s)  # pru-model can handle +- 268 mA

    sim.run(fn=current_trace, duration_s=duration_s)
    sim.plot(path_here, f"XP {config.name}, current discharge ramp (negative)")


def experiment_pulsed_discharge(config: VirtualStorageConfig) -> None:
    """Discharge virtual storage with a pulsed constant current."""
    dt_s = 0.2
    SoC_start = 1.0
    SoC_target = 0.0
    i_pulse = CurrentPulsed(
        I_pulse=-0.1, period_pulse=200, duration_pulse=100, SoC_target=SoC_target
    )  # pru-model can handle +- 268 mA
    sim = StorageSimulator(
        models=get_models(SoC_start, config, dt_s),
        dt_s=dt_s,
    )
    sim.run(fn=i_pulse.step, duration_s=min(1_000, duration_max))
    sim.plot(path_here, f"XP {config.name}, pulsed discharge .1A, 1000 s (figure_9a)")


def experiment_pulsed_charge(config: VirtualStorageConfig) -> None:
    """Charge virtual storage with a pulsed constant current."""
    dt_s = 0.2
    SoC_start = 0.0
    SoC_target = 1.0
    i_pulse = CurrentPulsed(
        I_pulse=0.1, period_pulse=200, duration_pulse=100, SoC_target=SoC_target
    )  # pru-model can handle +- 268 mA
    sim = StorageSimulator(
        models=get_models(SoC_start, config, dt_s),
        dt_s=dt_s,
    )
    sim.run(fn=i_pulse.step, duration_s=min(1_000, duration_max))
    sim.plot(path_here, f"XP {config.name}, pulsed charge .1A, 1000 s (figure_9b)")


def experiment_pulsed_resistive_charge(config: VirtualStorageConfig) -> None:
    """Charge virtual storage with a resistive constant voltage."""
    dt_s = 0.5
    SoC_start = 0.0
    i_pulse = ResistiveChargePulsed(R_Ohm=20, V_target=4.2, period_pulse=200, duration_pulse=100)
    sim = StorageSimulator(
        models=get_models(SoC_start, config, dt_s),
        dt_s=dt_s,
    )
    sim.run(fn=i_pulse.step, duration_s=min(3_000, duration_max))
    sim.plot(path_here, f"XP {config.name}, pulsed resistive charge 20 Ohm to 4.2 V, 3000 s")


def experiment_resistive_load(config: VirtualStorageConfig) -> None:
    """Charge virtual storage with a resistive constant voltage."""
    dt_s = 0.5
    SoC_start = 1.0

    def i_charge(_t_s: float, _s: float, V: float) -> float:
        return -V / 20

    sim = StorageSimulator(
        models=get_models(SoC_start, config, dt_s),
        dt_s=dt_s,
    )
    sim.run(fn=i_charge, duration_s=min(1_000, duration_max))
    sim.plot(path_here, f"XP {config.name}, resistive load 20 Ohm from 4.2 V, 1000 s")


def experiment_self_discharge() -> None:
    """Observe self-discharge behavior of virtual storage models."""
    dt_s = 0.2
    SoC_start = 1.0
    SoC_target = 0.9
    duration = timedelta(minutes=25)
    store = VirtualStorageConfig.capacitor(C_uF=100, V_rated=6.3)
    R_leak = store.calc_R_leak_capacitor(duration=duration, SoC_final=SoC_target, SoC_0=SoC_start)
    log.info("R_leak = %.2f Ohm", R_leak)
    config = VirtualStorageConfig.capacitor(C_uF=100, V_rated=6.3, R_leak_Ohm=R_leak)
    sim = StorageSimulator(
        models=get_models(SoC_start, config, dt_s),
        dt_s=dt_s,
    )

    def step(_t: float, _s: float, _v: float) -> float:
        return 0

    sim.run(fn=step, duration_s=min(duration.total_seconds(), duration_max))
    sim.plot(
        path_here,
        f"XP {config.name}, self-discharge, "
        f"SoC {SoC_start} to {SoC_target} in {duration.total_seconds()} s",
    )


if __name__ == "__main__":
    with multiprocessing.Pool() as pool:
        pool.apply_async(experiment_self_discharge)

        configs = [
            VirtualStorageConfig.capacitor(C_uF=10e6, V_rated=4.2),  # match charge with batteries
            VirtualStorageConfig.lipo(q_mAh=10),
            VirtualStorageConfig.lead_acid(q_mAh=10),
        ]

        for cfg in configs:
            pool.apply_async(experiment_pulsed_charge, args=(cfg,))
            pool.apply_async(experiment_pulsed_discharge, args=(cfg,))
            pool.apply_async(experiment_current_ramp_pos, args=(cfg,))
            pool.apply_async(experiment_current_ramp_neg, args=(cfg,))

        for cfg in configs[0:2]:
            pool.apply_async(experiment_pulsed_resistive_charge, args=(cfg,))
            pool.apply_async(experiment_resistive_load, args=(cfg,))

        pool.close()
        pool.join()



"""Separated data-type.

Done due to cyclic inheritance.
"""

from enum import Enum


class FirmwareDType(str, Enum):
    """Options for firmware-types."""

    base64_hex = "hex"
    base64_elf = "elf"
    path_hex = "path_hex"
    path_elf = "path_elf"



from pathlib import Path

import yaml


def load_yaml(file: str) -> dict:
    yaml_path = Path(__file__).resolve().with_name(file)
    with yaml_path.open() as _data:
        return yaml.safe_load(_data)


path_fwt = Path(__file__).parent.parent.resolve() / "fw_tools"
names_elf = ["build_msp.elf", "build_nrf.elf"]
files_elf = [path_fwt / name for name in names_elf]



import pytest
from pydantic import ValidationError
from shepherd_core.data_models.testbed import GPIO
from shepherd_core.data_models.testbed import MCU
from shepherd_core.data_models.testbed import Cape
from shepherd_core.data_models.testbed import Direction
from shepherd_core.data_models.testbed import Observer
from shepherd_core.data_models.testbed import ProgrammerProtocol
from shepherd_core.data_models.testbed import Target
from shepherd_core.data_models.testbed import Testbed as TasteBad

# ⤷ TasteBad avoids pytest-warning


def test_testbed_model_cape_min() -> None:
    cape = Cape(
        id=9999,
        name="cappi",
        version="1.0.0",
        description="lorem",
    )
    print(cape)


def test_testbed_model_gpio_min() -> None:
    gpio = GPIO(
        id=9999,
        name="gippi",
        reg_pru="ABCD",
        pin_pru="EFGH",
    )
    print(gpio)


def test_testbed_model_gpio_fault() -> None:
    with pytest.raises(ValidationError):
        GPIO(
            id=9999,
            name="gippi",
        )


def test_testbed_model_gpio_var() -> None:
    GPIO(
        id=9999,
        name="gippi",
        direction=Direction.Bidirectional,
        reg_pru="ABCD",
        pin_pru="EFGH",
    )


def test_testbed_model_mcu_min() -> None:
    mcu = MCU(
        id=9922,
        name="controller2",
        description="lorem",
        platform="arm32",
        core="STM32F7",
        prog_protocol=ProgrammerProtocol.SWD,
        fw_name_default="nananana",
    )
    print(mcu)


def test_testbed_model_observer_min() -> None:
    obs = Observer(
        id=9933,
        name="sheep120",
        description="not existing",
        ip="127.0.0.1",
        mac="FF:FF:FF:FF:FF:FF",
        room="IIE72",
        eth_port="375b2",
        cape=Cape(name="cape53"),
    )
    print(obs)


def test_testbed_model_observer_fault_cape_a() -> None:
    with pytest.raises(ValidationError):
        Observer(
            id=9933,
            name="sheep120",
            description="not existing",
            ip="127.0.0.1",
            mac="FF:FF:FF:FF:FF:FF",
            room="IIE72",
            eth_port="375b2",
            target_a=Target(id=3),
        )


def test_testbed_model_observer_fault_cape_b() -> None:
    with pytest.raises(ValidationError):
        Observer(
            id=9933,
            name="sheep120",
            description="not existing",
            ip="127.0.0.1",
            mac="FF:FF:FF:FF:FF:FF",
            room="IIE72",
            eth_port="375b2",
            target_b=Target(id=3),
        )


def test_testbed_model_observer_fault_target() -> None:
    obs = Observer(
        id=9933,
        name="sheep120",
        description="not existing",
        ip="127.0.0.1",
        mac="FF:FF:FF:FF:FF:FF",
        room="IIE72",
        eth_port="375b2",
        cape=Cape(name="cape53"),
        target_a=Target(id=3),
        target_b=Target(id=2),
    )
    obs.get_target(2)
    obs.get_target(3)
    obs.get_target_port(2)
    obs.get_target_port(3)
    with pytest.raises(KeyError):
        obs.get_target_port(123456)
    with pytest.raises(KeyError):
        obs.get_target(123456)


def test_testbed_model_target_min1() -> None:
    tgt = Target(
        id=9944,
        name="TerraTarget",
        version="v1.00",
        description="lorem",
        mcu1=MCU(name="MSP430FR"),
    )
    print(tgt)


def test_testbed_model_target_min2() -> None:
    Target(
        id=9944,
        name="TerraTarget",
        version="v1.00",
        description="lorem",
        mcu1="MSP430FR",
        mcu2="MSP430FR",
    )


def test_testbed_model_tb_min() -> None:
    tb = TasteBad(
        name="shepherd",
        id="9955",
        description="lorem",
        observers=[Observer(name="sheep02")],
        data_on_server="/mnt/driveA",
        data_on_observer="/mnt/driveB",
    )
    print(tb)


def test_testbed_model_tb_fault_observer() -> None:
    with pytest.raises(ValidationError):
        TasteBad(
            name="shepherd",
            id="9955",
            description="lorem",
            observers=[Observer(name="sheep02"), Observer(name="sheep02")],
            data_on_server="/mnt/driveA",
            data_on_observer="/mnt/driveB",
        )


def test_testbed_model_tb_fault_shared() -> None:
    with pytest.raises(ValidationError):
        TasteBad(
            name="shepherd",
            id="9955",
            description="lorem",
            observers=[Observer(name="sheep02")],
            data_on_server="/mnt/driveA",
            data_on_observer="/mnt/driveB",
            shared_storage=False,
        )



from datetime import timedelta
from pathlib import Path

from shepherd_core.data_models import EnergyEnvironment
from shepherd_core.data_models import Experiment
from shepherd_core.data_models import Firmware
from shepherd_core.data_models import TargetConfig
from shepherd_core.data_models import VirtualHarvesterConfig
from shepherd_core.data_models import VirtualSourceConfig
from shepherd_core.data_models.task import TestbedTasks as TasteBadTasks

from shepherd_core import local_now


def test_task_generation_file(tmp_path: Path) -> None:
    path = Path(__file__).with_name("example_config_experiment.yaml")
    xp1 = Experiment.from_file(path)
    tb_tasks = TasteBadTasks.from_xp(xp1)
    tb_tasks.to_file(tmp_path / "tbt1.yaml")


def test_task_generation_script(tmp_path: Path) -> None:
    hrv = VirtualHarvesterConfig(name="mppt_bq_thermoelectric")

    target_cfgs = [
        # first init similar to yaml
        TargetConfig(
            target_IDs=list(range(1, 4)),
            custom_IDs=list(range(3)),
            energy_env={"name": "SolarSunny"},
            virtual_source={"name": "diode+capacitor"},
            firmware1={"name": "nrf52_demo_rf"},
        ),
        # second Instance fully object-oriented
        TargetConfig(
            target_IDs=list(range(6, 9)),
            custom_IDs=list(range(7, 18)),
            energy_env=EnergyEnvironment(name="ThermoelectricWashingMachine"),
            virtual_source=VirtualSourceConfig(name="BQ25570-Schmitt", harvester=hrv),
            firmware1=Firmware(name="nrf52_demo_rf"),
            firmware2=Firmware(name="msp430_deep_sleep"),
        ),
    ]

    xperi = Experiment(
        name="meaningful Test-Name",
        time_start=local_now() + timedelta(minutes=30),
        target_configs=target_cfgs,
    )

    tb_tasks = TasteBadTasks.from_xp(xperi)
    tb_tasks.to_file(tmp_path / "tbt2.yaml")



"""Handling firmware for targets.

TODO: should be more generalized - currently only supports msp & nRF
"""

from pathlib import Path
from typing import Annotated
from typing import Any
from typing import TypedDict

from pydantic import StringConstraints
from pydantic import model_validator
from pydantic import validate_call
from typing_extensions import Self
from typing_extensions import Unpack

from shepherd_core import fw_tools
from shepherd_core.data_models.base.content import ContentModel
from shepherd_core.data_models.testbed.mcu import MCU
from shepherd_core.logger import log
from shepherd_core.testbed_client import tb_client

from .firmware_datatype import FirmwareDType

suffix_to_DType: dict = {
    # derived from wikipedia
    ".hex": FirmwareDType.base64_hex,
    ".ihex": FirmwareDType.base64_hex,
    ".ihx": FirmwareDType.base64_hex,
    ".elf": FirmwareDType.base64_elf,
    ".bin": FirmwareDType.base64_elf,
    ".o": FirmwareDType.base64_elf,
    ".out": FirmwareDType.base64_elf,
    ".so": FirmwareDType.base64_elf,
}

base64_to_path_DType: dict = {
    FirmwareDType.base64_hex: FirmwareDType.path_hex,
    FirmwareDType.base64_elf: FirmwareDType.path_elf,
}

arch_to_mcu: dict = {
    "em_msp430": {"name": "msp430fr"},
    "msp430": {"name": "msp430fr"},
    "arm": {"name": "nrf52"},
    "nrf52": {"name": "nrf52"},
}

FirmwareStr = Annotated[str, StringConstraints(min_length=3, max_length=8_000_000)]


class Firmware(ContentModel, title="Firmware of Target"):
    """meta-data representation of a data-component."""

    # General Metadata & Ownership -> ContentModel

    mcu: MCU

    data: FirmwareStr | Path
    data_type: FirmwareDType
    data_hash: str | None = None
    data_local: bool = True
    """ ⤷ signals that file has to be copied to testbed"""

    @model_validator(mode="before")
    @classmethod
    def query_database(cls, values: dict[str, Any]) -> dict[str, Any]:
        values, _ = tb_client.try_completing_model(cls.__name__, values)
        # crosscheck type with actual data
        _type = values.get("data_type")
        if _type in {
            FirmwareDType.base64_hex,
            FirmwareDType.base64_elf,
        }:
            try:
                _hash = fw_tools.base64_to_hash(values.get("data"))
            except ValueError:
                raise ValueError("Embedded Firmware seems to be faulty") from None
            if values.get("data_hash") is not None and _hash != values.get("data_hash"):
                raise ValueError("Embedded Firmware and Hash do not match!")
        elif _type in {
            FirmwareDType.path_hex,
            FirmwareDType.path_elf,
        }:
            try:
                _ = Path(values.get("data"))
            except (SyntaxError, NameError):
                raise ValueError("Firmware-Path is invalid") from None
        return tb_client.fill_in_user_data(values)

    @classmethod
    def from_firmware(
        cls,
        file: Path,
        *,
        embed: bool = True,
        **kwargs: Unpack[TypedDict],
    ) -> Self:
        """Embeds firmware and tries to fill parameters.

        ELF -> mcu und data_type are deducted
        HEX -> must supply mcu manually.
        """
        # TODO: use new determine_type() & determine_arch() and also allow to not embed
        kwargs["data_hash"] = fw_tools.file_to_hash(file)
        if embed:
            kwargs["data"] = fw_tools.file_to_base64(file)
            kwargs["data_local"] = False
        else:
            kwargs["data"] = Path(file).as_posix()
            kwargs["data_local"] = True

        if "data_type" not in kwargs:
            kwargs["data_type"] = suffix_to_DType[file.suffix.lower()]

        if kwargs["data_type"] == FirmwareDType.base64_hex:
            if fw_tools.is_hex_msp430(file):
                arch = "msp430"
            elif fw_tools.is_hex_nrf52(file):
                arch = "nrf52"
            else:
                raise ValueError("File is not a suitable HEX for the Testbed")
            if "mcu" not in kwargs:
                kwargs["mcu"] = arch_to_mcu[arch]

        # verification of ELF - warn if something is off
        # -> adds ARCH if it is able to derive
        if kwargs["data_type"] == FirmwareDType.base64_elf:
            arch = fw_tools.read_arch(file)
            try:
                if "msp430" in arch and not fw_tools.is_elf_msp430(file):
                    raise ValueError("File is not a ELF for msp430")
                if ("nrf52" in arch or "arm" in arch) and not fw_tools.is_elf_nrf52(file):
                    raise ValueError("File is not a ELF for nRF52")
            except RuntimeError:
                log.warning("ObjCopy not found -> Arch of Firmware can't be verified")
            log.debug("ELF-File '%s' has arch: %s", file.name, arch)
            if "mcu" not in kwargs:
                kwargs["mcu"] = arch_to_mcu[arch]

        if not embed:
            kwargs["data_type"] = base64_to_path_DType[kwargs["data_type"]]

        if "name" not in kwargs:
            kwargs["name"] = file.name
        return cls(**kwargs)

    def compare_hash(self, path: Path | None = None) -> bool:
        if self.data_hash is None:
            return True

        if path is not None and path.is_file():
            hash_new = fw_tools.file_to_hash(path)
            match = self.data_hash == hash_new
        else:
            hash_new = fw_tools.base64_to_hash(self.data)
            match = self.data_hash == hash_new

        if not match:
            log.warning("FW-Hash does not match with stored value!")
            # TODO: it might be more appropriate to raise here
        return match

    @validate_call
    def extract_firmware(self, file: Path) -> Path:
        """Store embedded fw-data in file.

        - file-suffix is derived from data-type and adapted
        - if provided path is a directory, the firmware-name is used
        """
        if file.is_dir():
            file = file / self.name
        file_new = fw_tools.extract_firmware(self.data, self.data_type, file)
        self.compare_hash(file_new)
        return file_new



"""Original KiBaM-Models with varying quality of detail."""

import math
import sys
from collections.abc import Callable

from pydantic import BaseModel
from pydantic import PositiveFloat
from pydantic import PositiveInt
from pydantic import validate_call
from typing_extensions import Self

from shepherd_core.data_models.content.virtual_storage_config import LuT_SIZE
from shepherd_core.data_models.content.virtual_storage_config import TIMESTEP_s_DEFAULT
from shepherd_core.data_models.content.virtual_storage_config import VirtualStorageConfig
from shepherd_core.data_models.content.virtual_storage_config import soc_t

from .virtual_storage_model import ModelStorage


class LUT(BaseModel):
    """Dynamic look-up table that can automatically be generated from a function."""

    x_min: float
    y_values: list[float]
    length: int
    interpolate: bool = False

    @classmethod
    @validate_call
    def generate(
        cls,
        x_min: PositiveFloat,
        y_fn: Callable,
        lut_size: PositiveInt = LuT_SIZE,
        *,
        optimize_clamp: bool = False,
        interpolate: bool = False,
    ) -> Self:
        """
        Generate a LUT with a specific width from a provided function.

        It has a minimum value, a size / width and a scale (linear / log2).
        y_fnc is a function that takes an argument and produces the lookup value.
        """
        if interpolate:
            # Note: dynamically creating .get() with setattr() was not successful
            optimize_clamp = False

        offset = 0.5 if optimize_clamp else 0
        x_values = [(i + offset) * x_min for i in range(lut_size)]
        y_values = [y_fn(x) for x in x_values]
        return cls(x_min=x_min, y_values=y_values, length=lut_size, interpolate=interpolate)

    def get(self, x_value: float) -> float:
        return self.get_interpol(x_value) if self.interpolate else self.get_discrete(x_value)

    def get_discrete(self, x_value: float) -> float:
        """Discrete LuT-lookup with typical stairs."""
        num = int(x_value / self.x_min)
        # ⤷ round() would be more appropriate, but in c/pru its just integer math
        idx = max(0, num)
        if idx >= self.length:  # len(self.y_values)
            idx = self.length - 1
        return self.y_values[idx]

    def get_interpol(self, x_value: float) -> float:
        """LuT-lookup with additional interpolation.

        Note: optimize-clamp must be disabled, otherwise this produces an offset
        """
        num = x_value / self.x_min
        if num <= 0:
            return self.y_values[0]
        if num >= self.length - 1:
            return self.y_values[self.length - 1]

        idx: int = math.floor(num)
        # high could be math.ceil(num), but also idx+1
        num_f: float = num - idx
        y_base = self.y_values[idx]
        y_delta = self.y_values[idx + 1] - y_base
        # TODO: y_delta[idx_l] could be a seconds LuT
        return y_base + y_delta * num_f


class ModelKiBaM(ModelStorage):
    """Naive implementation of the full hybrid KiBaM model from the paper.

    Introduced in "A Hybrid Battery Model Capable of Capturing Dynamic Circuit
    Characteristics and Nonlinear Capacity Effects".

    It is mostly focused on discharge, so it won't support

    - rate capacity effect and transients during charging
    - self discharge (as it was deemed too small)
    """

    @validate_call
    def __init__(
        self,
        cfg: VirtualStorageConfig,
        SoC_init: soc_t | None = None,
        dt_s: PositiveFloat = TIMESTEP_s_DEFAULT,
    ) -> None:
        # metadata for simulator
        self.cfg: VirtualStorageConfig = cfg
        self.dt_s: float = dt_s
        # state
        self.SoC: float = SoC_init if SoC_init is not None else cfg.SoC_init
        self.time_s: float = 0

        # Rate capacity effect
        self.C_unavailable: float = 0
        self.C_unavailable_last: float = 0

        # Transient tracking
        self.V_transient_S_max: float = 0
        self.V_transient_L_max: float = 0
        self.discharge_last: bool = False

        # Modified transient tracking
        self.V_transient_S: float = 0
        self.V_transient_L: float = 0

    def step(self, I_charge_A: float) -> tuple[float, float, float, float]:
        """Calculate the battery SoC & cell-voltage after drawing a current over a time-step."""
        # Step 1 verified separately using Figure 4
        # Steps 1 and 2 verified separately using Figure 10
        # Complete model verified using Figures 8 (a, b) and Figure 9 (a, b)
        I_cell = -I_charge_A

        # Step 0: Determine whether battery is charging or resting and
        #         calculate time since last switch
        if self.discharge_last != (I_cell > 0):  # Reset time delta when current sign changes
            self.discharge_last = I_cell > 0
            self.time_s = 0
            self.C_unavailable_last = self.C_unavailable
            # ⤷ Save C_unavailable at time of switch

        self.time_s += self.dt_s
        # ⤷ Consider time delta including this iteration (we want v_trans after the current step)

        # Step 1: Calculate unavailable capacity after dt
        #         (due to rate capacity and recovery effect) (equation 17)
        # Note: it seems possible to remove the 2nd branch if
        #       charging is considered (see Plus-Model)
        if I_cell > 0:  # Discharging
            self.C_unavailable = (
                self.C_unavailable_last * math.pow(math.e, -self.cfg.kdash * self.time_s)
                + (1 - self.cfg.p_rce)
                * I_cell
                / self.cfg.p_rce
                * (1 - math.pow(math.e, -self.cfg.kdash * self.time_s))
                / self.cfg.kdash
            )
        else:  # Recovering
            self.C_unavailable = self.C_unavailable_last * math.pow(
                math.e, -self.cfg.kdash * self.time_s
            )

        # Step 2: Calculate SoC after dt (equation 6; modified for discrete operation)
        # ⤷ MODIFIED: clamp both SoC to 0..1
        self.SoC = self.SoC - 1 / self.cfg.q_As * (I_cell * self.dt_s)
        self.SoC = min(max(self.SoC, 0.0), 1.0)
        SoC_eff = self.SoC - 1 / self.cfg.q_As * self.C_unavailable
        SoC_eff = max(SoC_eff, 0.0)

        # Step 3: Calculate V_OC after dt (equation 7)
        V_OC = self.cfg.calc_V_OC(SoC_eff)

        # Step 4: Calculate resistance and capacitance values after dt (equation 12)
        R_series = self.cfg.calc_R_series(SoC_eff)
        R_transient_S = self.cfg.calc_R_transient_S(SoC_eff)
        C_transient_S = self.cfg.calc_C_transient_S(SoC_eff)
        R_transient_L = self.cfg.calc_R_transient_L(SoC_eff)
        C_transient_L = self.cfg.calc_C_transient_L(SoC_eff)

        # Step 5: Calculate transient voltages (equations 10 and 11)
        # ⤷ MODIFIED: prevent both tau_X from becoming 0
        tau_S = max(R_transient_S * C_transient_S, sys.float_info.min)
        if I_cell > 0:  # Discharging
            V_transient_S = R_transient_S * I_cell * (1 - math.pow(math.e, -self.time_s / tau_S))
            self.V_transient_S_max = V_transient_S
        else:  # Recovering
            V_transient_S = self.V_transient_S_max * math.pow(math.e, -self.time_s / tau_S)

        tau_L = max(R_transient_L * C_transient_L, sys.float_info.min)
        if I_cell > 0:  # Discharging
            V_transient_L = R_transient_L * I_cell * (1 - math.pow(math.e, -self.time_s / tau_L))
            self.V_transient_L_max = V_transient_L
        else:  # Recovering
            V_transient_L = self.V_transient_L_max * math.pow(math.e, -self.time_s / tau_L)

        # Step 6: Calculate cell voltage (equations 8 and 9)
        # ⤷ MODIFIED: limit V_cell to >=0
        V_transient = V_transient_S + V_transient_L
        V_cell = V_OC - I_cell * R_series - V_transient
        V_cell = max(V_cell, 0)

        return V_OC, V_cell, self.SoC, SoC_eff


class ModelKiBaMPlus(ModelStorage):
    """Hybrid KiBaM model from the paper with certain extensions.

    Extended by [@jonkub](https://github.com/jonkub) with streamlined math.

    Modifications:

    1. support rate capacity during charging (Step 1)
    2. support transient tracking during charging (Step 5)
    3. support self discharge (step 2a) via a parallel leakage resistor
    4. support signaling 0 % SoC by nulling voltage
    """

    @validate_call
    def __init__(
        self,
        cfg: VirtualStorageConfig,
        SoC_init: soc_t | None = None,
        dt_s: PositiveFloat = TIMESTEP_s_DEFAULT,
    ) -> None:
        # metadata for simulator
        self.cfg: VirtualStorageConfig = cfg
        self.dt_s: float = dt_s
        # state
        self.SoC: float = SoC_init if SoC_init is not None else cfg.SoC_init
        self.time_s: float = 0

        # Rate capacity effect
        self.C_unavailable: float = 0
        self.C_unavailable_last: float = 0

        # Transient tracking
        self.discharge_last: bool = False

        # Modified transient tracking
        self.V_transient_S: float = 0
        self.V_transient_L: float = 0

    def step(self, I_charge_A: float) -> tuple[float, float, float, float]:
        """Calculate the battery SoC & cell-voltage after drawing a current over a time-step.

        - Step 1 verified separately using Figure 4
        - Steps 1 and 2 verified separately using Figure 10
        - Complete model verified using Figures 8 (a, b) and Figure 9 (a, b)
        """
        I_cell = -I_charge_A

        # Step 0: Determine whether battery is charging or resting and
        #         calculate time since last switch
        if self.discharge_last != (I_cell > 0):  # Reset time delta when current sign changes
            self.discharge_last = I_cell > 0
            self.time_s = 0
            self.C_unavailable_last = self.C_unavailable  # Save C_unavailable at time of switch

        self.time_s += self.dt_s
        # ⤷ Consider time delta including this iteration (we want v_trans after the current step)

        # Step 1: Calculate unavailable capacity after dt
        #         (due to rate capacity and recovery effect) (equation 17)
        # TODO: if this should be used in production, additional verification is required
        #      (analytically derive versions of eq. 16/17 without time range restrictions)
        #       parameters for rate effect could only be valid for discharge
        #       Note: other paper has charging-curves (fig9b) - could be used for verification
        self.C_unavailable = (
            self.C_unavailable_last * math.pow(math.e, -self.cfg.kdash * self.time_s)
            + (1 - self.cfg.p_rce)
            * I_cell
            / self.cfg.p_rce
            * (1 - math.pow(math.e, -self.cfg.kdash * self.time_s))
            / self.cfg.kdash
        )

        # Step 2a: Calculate and add self-discharge current to SoC-Eq. below
        I_leak = self.cfg.calc_V_OC(self.SoC) / self.cfg.R_leak_Ohm

        # Step 2: Calculate SoC after dt (equation 6; modified for discrete operation)
        # ⤷ MODIFIED: clamp both SoC to 0..1
        self.SoC = self.SoC - (I_cell + I_leak) * self.dt_s / self.cfg.q_As
        self.SoC = min(max(self.SoC, 0.0), 1.0)
        SoC_eff = self.SoC - 1 / self.cfg.q_As * self.C_unavailable
        SoC_eff = min(max(SoC_eff, 0.0), 1.0)
        # ⤷ Note: limiting SoC_eff to <=1 should NOT be needed, but
        #         C_unavailable can become negative during charging (see assumption in step1).

        # Step 3: Calculate V_OC after dt (equation 7)
        V_OC = self.cfg.calc_V_OC(SoC_eff)

        # Step 4: Calculate resistance and capacitance values after dt (equation 12)
        R_series = self.cfg.calc_R_series(SoC_eff)
        R_transient_S = self.cfg.calc_R_transient_S(SoC_eff)
        C_transient_S = self.cfg.calc_C_transient_S(SoC_eff)
        R_transient_L = self.cfg.calc_R_transient_L(SoC_eff)
        C_transient_L = self.cfg.calc_C_transient_L(SoC_eff)

        # Step 5: Calculate transient voltages (equations 10 and 11)
        # ⤷ MODIFIED: prevent both tau_X from becoming 0
        tau_S = max(R_transient_S * C_transient_S, sys.float_info.min)
        tau_L = max(R_transient_L * C_transient_L, sys.float_info.min)
        self.V_transient_S = R_transient_S * I_cell + (
            self.V_transient_S - R_transient_S * I_cell
        ) * math.pow(math.e, -self.dt_s / tau_S)
        self.V_transient_L = R_transient_L * I_cell + (
            self.V_transient_L - R_transient_L * I_cell
        ) * math.pow(math.e, -self.dt_s / tau_L)

        # Step 6: Calculate cell voltage (equations 8 and 9)
        # ⤷ MODIFIED: limit V_cell to >=0
        V_transient = self.V_transient_S + self.V_transient_L
        V_cell = V_OC - I_cell * R_series - V_transient
        V_cell = max(V_cell, 0)
        if self.SoC == 0:
            V_cell = 0  # make sure no energy can be extracted when empty

        return V_OC, V_cell, self.SoC, SoC_eff


class ModelKiBaMSimple(ModelStorage):
    """PRU-optimized model with a set of simplifications.

    Modifications by [@jonkub](https://github.com/jonkub):

    - omit transient voltages (step 4 & 5, expensive calculation)
    - omit rate capacity effect (step 1, expensive calculation)
    - replace two expensive Fn by LuT
        - mapping SoC to open circuit voltage (step 3)
        - mapping SoC to series resistance (step 4)
    - add self discharge resistance (step 2a)
    - support signaling 0 % SoC by nulling voltage

    Compared to the current shepherd capacitor (charge-based), it:

    - supports emulation of battery types like lipo and lead acid (non-linear SOC-to-V_OC mapping)
    - has a parallel leakage resistor instead of an oversimplified leakage current
    - a series resistance is added to improve model matching
    - as a drawback the open circuit voltage is quantified and shows steps (LuT with 128 entries)

    """

    @validate_call
    def __init__(
        self,
        cfg: VirtualStorageConfig,
        SoC_init: soc_t | None = None,
        dt_s: PositiveFloat = TIMESTEP_s_DEFAULT,
        *,
        optimize_clamp: bool = False,
        interpolate: bool = False,
    ) -> None:
        # metadata for simulator
        self.cfg: VirtualStorageConfig = cfg
        self.dt_s = dt_s
        # pre-calculate constants
        self.V_OC_LuT: LUT = LUT.generate(
            1.0 / LuT_SIZE,
            y_fn=cfg.calc_V_OC,
            lut_size=LuT_SIZE,
            optimize_clamp=optimize_clamp,
            interpolate=interpolate,
        )
        self.R_series_LuT: LUT = LUT.generate(
            1.0 / LuT_SIZE,
            y_fn=cfg.calc_R_series,
            lut_size=LuT_SIZE,
            optimize_clamp=optimize_clamp,
            interpolate=interpolate,
        )
        self.Constant_s_per_As: float = dt_s / cfg.q_As
        self.Constant_1_per_Ohm: float = 1.0 / cfg.R_leak_Ohm
        # state
        self.SoC: float = SoC_init if SoC_init is not None else cfg.SoC_init

    def step(self, I_charge_A: float) -> tuple[float, float, float, float]:
        """Calculate the battery SoC & cell-voltage after drawing a current over a time-step."""
        I_cell = -I_charge_A
        # Step 2a: Calculate self-discharge (drainage)
        I_leak = self.V_OC_LuT.get(self.SoC) * self.Constant_1_per_Ohm

        # Step 2: Calculate SoC after dt (equation 6; modified for discrete operation)
        #       = SoC - 1 / C * (i_cell * dt)
        self.SoC = self.SoC - (I_cell + I_leak) * self.Constant_s_per_As
        SoC_eff = self.SoC = min(max(self.SoC, 0.0), 1.0)
        # ⤷ MODIFIED: removed term due to omission of rate capacity effect
        # ⤷ MODIFIED: clamp SoC to 0..1

        # Step 3: Calculate V_OC after dt (equation 7)
        # MODIFIED to use a lookup table instead
        V_OC = self.V_OC_LuT.get(SoC_eff)

        # Step 4: Calculate resistance and capacitance values after dt (equation 12)
        # MODIFIED: removed terms due to omission of transient voltages
        # MODIFIED to use a lookup table instead
        R_series = self.R_series_LuT.get(SoC_eff)

        # Step 5: Calculate transient voltages (equations 10 and 11)
        # MODIFIED: removed due to omission of transient voltages

        # Step 6: Calculate cell voltage (equations 8 and 9)
        # MODIFIED: removed term due to omission of transient voltages
        # MODIFIED: limit V_cell to >=0
        V_cell = V_OC - I_cell * R_series
        V_cell = max(V_cell, 0.0)
        if self.SoC == 0:
            V_cell = 0  # make sure no energy can be extracted when empty

        return V_OC, V_cell, self.SoC, SoC_eff


class ModelShpCap(ModelStorage):
    """A derived model from shepherd-codebase for comparing to KiBaM-capacitor.

    This model was used for the intermediate storage capacitor until
    the battery-model was implemented.
    """

    @validate_call
    def __init__(
        self,
        cfg: VirtualStorageConfig,
        SoC_init: soc_t | None = None,
        dt_s: PositiveFloat = TIMESTEP_s_DEFAULT,
    ) -> None:
        # metadata for simulator
        self.cfg: VirtualStorageConfig = cfg
        self.dt_s = dt_s
        # pre-calculate constants
        self.V_mid_max_V = cfg.calc_V_OC(1.0)
        C_mid_uF = 1e6 * cfg.q_As / self.V_mid_max_V
        C_mid_uF = max(C_mid_uF, 0.001)
        SAMPLERATE_SPS = 1.0 / dt_s
        self.Constant_s_per_F = 1e6 / (C_mid_uF * SAMPLERATE_SPS)
        self.Constant_1_per_Ohm: float = 1.0 / cfg.R_leak_Ohm
        # state
        SoC_init = SoC_init if SoC_init is not None else cfg.SoC_init
        self.V_mid_V = cfg.calc_V_OC(SoC_init)

    def step(self, I_charge_A: float) -> tuple[float, float, float, float]:
        # in PRU P_inp and P_out are calculated and combined to determine current
        # similar to: P_sum_W = P_inp_W - P_out_W, I_mid_A = P_sum_W / V_mid_V
        I_mid_A = I_charge_A - self.V_mid_V * self.Constant_1_per_Ohm
        dV_mid_V = I_mid_A * self.Constant_s_per_F
        self.V_mid_V += dV_mid_V

        self.V_mid_V = min(self.V_mid_V, self.V_mid_max_V)
        self.V_mid_V = max(self.V_mid_V, sys.float_info.min)
        SoC = self.V_mid_V / self.V_mid_max_V
        return self.V_mid_V, self.V_mid_V, SoC, SoC



"""Client-Class to access a testbed instance over the web."""

from importlib import import_module
from pathlib import Path
from typing import Any

from pydantic import validate_call

from shepherd_core.config import config
from shepherd_core.data_models.base.shepherd import ShpModel
from shepherd_core.data_models.base.wrapper import Wrapper

from .client_abc_fix import AbcClient
from .user_model import User


class WebClient(AbcClient):
    """Client-Class to access a testbed instance over the web.

    For online-queries the lib can be connected to the testbed-server.
    NOTE: there are 3 states:
    - unconnected -> demo-fixtures are queried (locally)
    - connected -> publicly available data is queried online
    - logged in with valid token -> also private data is queried online
    """

    testbed_server_default = "https://shepherd.cfaed.tu-dresden.de:8000/testbed"

    def __init__(self, server: str | None = None, token: str | Path | None = None) -> None:
        """Connect to Testbed-Server with optional token and server-address.

        server: optional address to shepherd-server-endpoint
        token: your account validation. if omitted, only public data is available
        """
        super().__init__()
        if not hasattr(self, "_token"):
            # add default values
            self._token: str = "basic_public_access"  # noqa: S105
            self._server: str = config.TESTBED_SERVER
            self._user: User | None = None
            self._key: str | None = None
            self._connected: bool = False
            self._req = None

        if not self._connected:
            self._connect(server, token)

    # ABC Functions below

    def insert(self, data: ShpModel) -> bool:
        if self._req is None:
            return False
        wrap = Wrapper(
            datatype=type(data).__name__,
            parameters=data.model_dump(),
        )
        r = self._req.post(self._server + "/add", data=wrap.model_dump_json(), timeout=2)
        r.raise_for_status()
        return True

    def query_ids(self, model_type: str) -> list[int]:
        raise NotImplementedError("TODO")

    def query_names(self, model_type: str) -> list[str]:
        raise NotImplementedError("TODO")

    def query_item(self, model_type: str, uid: int | None = None, name: str | None = None) -> dict:
        raise NotImplementedError("TODO")

    def try_inheritance(
        self, model_type: str, values: dict[str, Any]
    ) -> tuple[dict[str, Any], list[str]]:
        raise NotImplementedError("TODO")

    def fill_in_user_data(self, values: dict[str, Any]) -> dict[str, Any]:
        if self._user is None:
            return values
        if values.get("owner") is None:
            values["owner"] = self._user.name
        if values.get("group") is None:
            values["group"] = self._user.group
        return values

    # Below are extra FNs not in ABC

    @validate_call
    def _connect(self, server: str | None = None, token: str | Path | None = None) -> bool:
        """Establish connection to testbed-server.

        TODO: totally not finished
        """
        if isinstance(token, Path):
            if not token.exists():
                raise FileNotFoundError("Token-Path does not exist")
            with token.resolve().open() as file:
                self._token = file.read()
        elif isinstance(token, str):
            self._token = self._token

        if isinstance(server, str):
            self._server = server.lower()

        self._req = import_module("requests")  # here due to slow startup

        # extended connection-test:
        self._query_session_key()
        self._connected = True
        return self._query_user_data()

    def _query_session_key(self) -> bool:
        if self._server and self._req is not None:
            r = self._req.get(self._server + "/session_key", timeout=2)
            r.raise_for_status()
            self._key = r.json()["value"]  # TODO: not finished
            return True
        return False

    def _query_user_data(self) -> bool:
        if self._server and self._req is not None:
            r = self._req.get(self._server + "/user?token=" + self._token, timeout=2)
            # TODO: possibly a security nightmare (send via json or encrypted via public key?)
            r.raise_for_status()
            self._user = User(**r.json())
            return True
        return False

    def submit_experiment(self, xp: ShpModel) -> str:
        """Transmit XP to server to validate its feasibility.

        - Experiment will be added to DB (if not present)
            - if the same experiment is resubmitted it will just return the ID of that XP
        - Experiment will be validated by converting it into a task-set (additional validation)
        - optional: the scheduler should validate there are no time-collisions

        Will return an ID if valid, otherwise an empty string.
        TODO: maybe its better to throw specific errors if validation fails
        TODO: is it better to include these experiment-related FNs in Xp-Class?
        TODO: Experiment-typehint for argument triggers circular import
        """
        raise NotImplementedError("TODO")

    def schedule_experiment(self, id_xp: str) -> bool:
        """Enqueue XP on testbed."""
        raise NotImplementedError("TODO")

    def get_experiment_status(self, id_xp: str) -> str:
        """Ask server about current state of XP.

        - after valid submission: disabled / deactivated
        - after scheduling: scheduled
        - before start-time: preparing
        - during run: active
        - after run: post-processing (collecting & assembling data)
        - finished: ready to download
        """
        raise NotImplementedError("TODO")

    def get_experiment_results(self, id_xp: str, path: Path) -> bool:
        """Download resulting files."""
        raise NotImplementedError("TODO")



"""Helper-functions for handling firmware around the testbed."""

from typing import Any

try:
    from importlib.util import find_spec

    find_spec("pwnlib.elf.ELF")
except ImportError:
    # replace missing dependencies from elf-only pwntools
    import sys
    from importlib.util import find_spec
    from unittest.mock import MagicMock

    class Mock(MagicMock):
        @classmethod
        def __getattr__(cls, name: str) -> Any:
            return MagicMock()

    MOCK_MODULES = [
        "socks",  # general lib-init / context-init
        "serial",
        # troublemaker on bbone and/or windows
        "capstone",
        "paramiko",
        "unicorn",
        "cffi",
    ]
    # only update when module is not avail
    MOCK_MODULES = [mod_name for mod_name in MOCK_MODULES if find_spec(mod_name) is None]
    sys.modules.update((mod_name, Mock()) for mod_name in MOCK_MODULES)

from .converter import base64_to_file
from .converter import base64_to_hash
from .converter import extract_firmware
from .converter import file_to_base64
from .converter import file_to_hash
from .converter import firmware_to_hex
from .converter_elf import elf_to_hex
from .patcher import find_symbol
from .patcher import modify_symbol_value
from .patcher import modify_uid
from .patcher import read_arch
from .patcher import read_symbol
from .patcher import read_uid
from .validation import determine_arch
from .validation import determine_type
from .validation import is_elf
from .validation import is_elf_msp430
from .validation import is_elf_nrf52
from .validation import is_hex
from .validation import is_hex_msp430
from .validation import is_hex_nrf52

__all__ = [
    "base64_to_file",
    "base64_to_hash",
    "determine_arch",
    "determine_type",
    "elf_to_hex",
    "extract_firmware",
    "file_to_base64",
    "file_to_hash",
    "find_symbol",
    "firmware_to_hex",
    "is_elf",
    "is_elf_msp430",
    "is_elf_nrf52",
    "is_hex",
    "is_hex_msp430",
    "is_hex_nrf52",
    "modify_symbol_value",
    "modify_uid",
    "read_arch",
    "read_symbol",
    "read_uid",
]
